"""
pyrad.prod.process_spectra_products
===================================

Functions for obtaining Pyrad products from spectra datasets

.. autosummary::
    :toctree: generated/

    generate_spectra_products

"""

from warnings import warn
from copy import deepcopy

import numpy as np

import pyart
from pyart.util import datetime_from_radar

from ..io.io_aux import get_fieldname_pyart
from ..io.io_aux import get_save_dir, make_filename

from ..graph.plots_spectra import plot_range_Doppler, plot_Doppler
from ..graph.plots_spectra import plot_complex_range_Doppler
from ..graph.plots_spectra import plot_amp_phase_range_Doppler
from ..graph.plots_spectra import plot_complex_Doppler, plot_amp_phase_Doppler
from ..graph.plots_spectra import plot_time_Doppler, plot_complex_time_Doppler
from ..graph.plots_spectra import plot_amp_phase_time_Doppler
from ..graph.plots_spectra import plot_angle_Doppler
from ..graph.plots_spectra import plot_complex_angle_Doppler
from ..graph.plots_spectra import plot_amp_phase_angle_Doppler

from ..util.radar_utils import find_ray_index, find_rng_index


def generate_spectra_products(dataset, prdcfg):
    """
    generates spectra products. Accepted product types:
        'AMPLITUDE_PHASE_ANGLE_DOPPLER': Makes an angle Doppler plot of
            complex spectra or IQ data. The plot can be along azimuth or along
            range. It is plotted separately the module and the phase of the
            signal.
            User defined parameters:
                voltype: str
                    name of the pyrad variable to use, it must be available in the dataset
                along_azi : bool
                    If true the plot is performed along azimuth, otherwise
                    along elevation. Default true
                ang : float
                    The fixed angle (deg). Default 0.
                rng : float
                    The fixed range (m). Default 0.
                ang_tol : float
                    The fixed angle tolerance (deg). Default 1.
                rng_tol : float
                    The fixed range tolerance (m). Default 50.
                xaxis_info : str
                    The xaxis type. Can be 'Doppler_velocity',
                    'Doppler_frequency' or 'pulse_number'
                ampli_vmin, ampli_vmax, phase_vmin, phase_vmax : float or None
                    Minimum and maximum of the color scale for the module and
                    phase
        'AMPLITUDE_PHASE_DOPPLER': Plots a complex Doppler spectrum or IQ data
            making two separate plots for the module and phase of the signal
            User defined parameters:
                voltype: str
                    name of the pyrad variable to use, it must be available in the dataset
                azi : float
                    azimuth (deg) of the ray to plot
                ele : float
                    elevation (deg) of the ray to plot
                rng : float
                   range (m) of the ray to plot
                azi_tol : float
                    azimuth (deg) tolerance respect to nominal position to plot.
                    Default 1.
                ele_tol : float
                    elevation (deg) tolerance respect to nominal position to plot.
                    Default 1.
                rng_tol : float
                    elevation (deg) tolerance respect to nominal position to plot.
                    Default 50.
                ind_ray : int
                    index of the ray to plot. Alternative to
                    defining its antenna coordinates
                ind_rng : int
                    index of the range to plot. Alternative to
                    defining its antenna coordinates
                xaxis_info : str
                    The xaxis type. Can be 'Doppler_velocity',
                    'Doppler_frequency' or 'pulse_number'
                ampli_vmin, ampli_vmax, phase_vmin, phase_vmax : float or None
                    Minimum and maximum of the color scale for the module and
                    phase
        'AMPLITUDE_PHASE_RANGE_DOPPLER': Plots a complex spectra or IQ data
            range-Doppler making two separate plots for the module and phase
            of the signal User defined parameters:
                voltype: str
                    name of the pyrad variable to use, it must be available in the dataset
                azi : float
                    azimuth (deg) of the ray to plot
                ele : float
                    elevation (deg) of the ray to plot
                azi_tol : float
                    azimuth (deg) tolerance respect to nominal position to plot.
                    Default 1.
                ele_tol : float
                    elevation (deg) tolerance respect to nominal position to plot.
                    Default 1.
                ind_ray : int
                    index of the ray to plot. Alternative to
                    defining its antenna coordinates
                xaxis_info : str
                    The xaxis type. Can be 'Doppler_velocity',
                    'Doppler_frequency' or 'pulse_number'
                ampli_vmin, ampli_vmax, phase_vmin, phase_vmax : float or None
                    Minimum and maximum of the color scale for the module and
                    phase
        'AMPLITUDE_PHASE_TIME_DOPPLER': Plots a complex spectra or IQ data
            time-Doppler making two separate plots for the module and phase of
            the signal
            User defined parameters:
                voltype: str
                    name of the pyrad variable to use, it must be available in the dataset
                antenna_coordinates_az_el_r : list of float
                    azimuth, elevation and range coordinates of the reference point (3 elements)
                point_coordinates_WGS84_lon_lat_alt : list of float
                    lon, lat (deg in WGS84 reference) and alt (in m) coordinates of the reference point.
                    Will be used only if antenna_coordinates_az_el_r is not provided.
                xaxis_info : str
                    The xaxis type. Can be 'Doppler_velocity' or
                    'Doppler frequency'
                ampli_vmin : float or None
                    Minimum of the color scale for the module
                ampli_vmax : float or None
                    Maximum of the color scale for the module
                phase_vmin : float or None
                    Minimum of the color scale for the module
                phase_vmax : float or None
                    Minimum of the color scale for the phase
                plot_type : str
                    Can be 'final' or 'temporal'. If final the data is only
                    plotted at the end of the processing
        'ANGLE_DOPPLER': Makes an angle Doppler plot. The plot can be along
            azimuth or along range
            User defined parameters:
                voltype: str
                    name of the pyrad variable to use, it must be available in the dataset
                along_azi : bool
                    If true the plot is performed along azimuth, otherwise
                    along elevation. Default true
                ang : float
                    The fixed angle (deg). Default 0.
                rng : float
                    The fixed range (m). Default 0.
                ang_tol : float
                    The fixed angle tolerance (deg). Default 1.
                rng_tol : float
                    The fixed range tolerance (m). Default 50.
                xaxis_info : str
                    The xaxis type. Can be 'Doppler_velocity',
                    'Doppler_frequency' or 'pulse_number'
                vmin : float or None
                    Minimum of the color scale
                vmax : float or None
                    Maximum of the color scale
        'COMPLEX_ANGLE_DOPPLER': Makes an angle Doppler plot of complex
            spectra or IQ data. The plot can be along azimuth or along range.
            The real and imaginary parts are plotted separately
            User defined parameters:
                voltype: str
                    name of the pyrad variable to use, it must be available in the dataset
                along_azi : bool
                    If true the plot is performed along azimuth, otherwise
                    along elevation. Default true
                ang : float
                    The fixed angle (deg). Default 0.
                rng : float
                    The fixed range (m). Default 0.
                ang_tol : float
                    The fixed angle tolerance (deg). Default 1.
                rng_tol : float
                    The fixed range tolerance (m). Default 50.
                xaxis_info : str
                    The xaxis type. Can be 'Doppler_velocity',
                    'Doppler_frequency' or 'pulse_number'
                vmin : float or None
                    Minimum of the color scale
                vmax : float or None
                    Maximum of the color scale
        'COMPLEX_DOPPLER': Plots a complex Doppler spectrum or IQ data making
            two separate plots for the real and imaginary parts
            User defined parameters:
                voltype: str
                    name of the pyrad variable to use, it must be available in the dataset
                azi : float
                    azimuth (deg) of the ray to plot
                ele : float
                    elevation (deg) of the ray to plot
                rng : float
                   range (m) of the ray to plot
                azi_tol : float
                    azimuth (deg) tolerance respect to nominal position to plot.
                    Default 1.
                ele_tol : float
                    elevation (deg) tolerance respect to nominal position to plot.
                    Default 1.
                rng_tol : float
                    elevation (deg) tolerance respect to nominal position to plot.
                    Default 50.
                ind_ray : int
                    index of the ray to plot. Alternative to
                    defining its antenna coordinates
                ind_rng : int
                    index of the range to plot. Alternative to
                    defining its antenna coordinates
                xaxis_info : str
                    The xaxis type. Can be 'Doppler_velocity',
                    'Doppler_frequency' or 'pulse_number'
                vmin : float or None
                    Minimum of the color scale
                vmax : float or None
                    Maximum of the color scale
        'COMPLEX_RANGE_DOPPLER': Plots the complex spectra or IQ data
            range-Doppler making two separate plots for the real and imaginary
            parts
            User defined parameters:
                voltype: str
                    name of the pyrad variable to use, it must be available in the dataset
                azi : float
                    azimuth (deg) of the ray to plot
                ele : float
                    elevation (deg) of the ray to plot
                rng : float
                   range (m) of the ray to plot
                azi_tol : float
                    azimuth (deg) tolerance respect to nominal position to plot.
                    Default 1.
                ele_tol : float
                    elevation (deg) tolerance respect to nominal position to plot.
                    Default 1.
                rng_tol : float
                    elevation (deg) tolerance respect to nominal position to plot.
                    Default 50.
                ind_ray : int
                    index of the ray to plot. Alternative to
                    defining its antenna coordinates
                xaxis_info : str
                    The xaxis type. Can be 'Doppler_velocity',
                    'Doppler_frequency' or 'pulse_number'
                vmin : float or None
                    Minimum of the color scale
                vmax : float or None
                    Maximum of the color scale
        'COMPLEX_TIME_DOPPLER': Plots the complex spectra or IQ data
            time-Doppler making two separate plots for the real and imaginary
            parts
            User defined parameters:
                voltype: str
                    name of the pyrad variable to use, it must be available in the dataset
                antenna_coordinates_az_el_r : list of float
                    azimuth, elevation and range coordinates of the reference point (3 elements)
                point_coordinates_WGS84_lon_lat_alt : list of float
                    lon, lat (deg in WGS84 reference) and alt (in m) coordinates of the reference point.
                    Will be used only if antenna_coordinates_az_el_r is not provided.
                xaxis_info : str
                    The xaxis type. Can be 'Doppler_velocity' or
                    'Doppler frequency'
                vmin : float or None
                    Minimum of the color scale
                vmax : float or None
                    Maximum of the color scale
                plot_type : str
                    Can be 'final' or 'temporal'. If final the data is only
                    plotted at the end of the processing
        'DOPPLER': Plots a Doppler spectrum variable or IQ data variable
            User defined parameters:
                voltype: str
                    name of the pyrad variable to use, it must be available in the dataset
                azi : float
                    azimuth (deg) of the ray to plot
                ele : float
                    elevation (deg) of the ray to plot
                rng : float
                   range (m) of the ray to plot
                azi_tol : float
                    azimuth (deg) tolerance respect to nominal position to plot.
                    Default 1.
                ele_tol : float
                    elevation (deg) tolerance respect to nominal position to plot.
                    Default 1.
                rng_tol : float
                    elevation (deg) tolerance respect to nominal position to plot.
                    Default 50.
                ind_ray : int
                    index of the ray to plot. Alternative to
                    defining its antenna coordinates
                xaxis_info : str
                    The xaxis type. Can be 'Doppler_velocity',
                    'Doppler_frequency' or 'pulse_number'
                vmin : float or None
                    Minimum of the color scale
                vmax : float or None
                    Maximum of the color scale
        'RANGE_DOPPLER': Makes a range-Doppler plot of spectral or IQ data
            User defined parameters:
                voltype: str
                    name of the pyrad variable to use, it must be available in the dataset
                azi : float
                    azimuth (deg) of the ray to plot
                ele : float
                    elevation (deg) of the ray to plot
                azi_tol : float
                    azimuth (deg) tolerance respect to nominal position to plot.
                    Default 1.
                ele_tol : float
                    elevation (deg) tolerance respect to nominal position to plot.
                    Default 1.
                ind_ray : int
                    index of the ray to plot. Alternative to
                    defining its antenna coordinates
                xaxis_info : str
                    The xaxis type. Can be 'Doppler_velocity',
                    'Doppler_frequency' or 'pulse_number'
                vmin : float or None
                    Minimum of the color scale
                vmax : float or None
                    Maximum of the color scale
        'SAVEALL': Saves radar spectra or IQ volume data including all or a
            list of userdefined fields in a netcdf file
            User defined parameters:
                datatypes: list of str or None
                    The list of data types to save. If it is None, all fields
                    in the radar object will be saved
                physical: Bool
                    If True the data will be saved in physical units (floats).
                    Otherwise it will be quantized and saved as binary
        'SAVEVOL': Saves one field of a radar spectra or IQ volume data in a
            netcdf file
            User defined parameters:
                voltype: str
                    name of the pyrad variable to use, it must be available in the dataset
                physical: Bool
                    If True the data will be saved in physical units (floats).
                    Otherwise it will be quantized and saved as binary
        'TIME_DOPPLER': Makes a time-Doppler plot of spectral or IQ data at a
            point of interest.
            User defined parameters:
                voltype: str
                    name of the pyrad variable to use, it must be available in the dataset
                antenna_coordinates_az_el_r : list of float
                    azimuth, elevation and range coordinates of the reference point (3 elements)
                point_coordinates_WGS84_lon_lat_alt : list of float
                    lon, lat (deg in WGS84 reference) and alt (in m) coordinates of the reference point.
                    Will be used only if antenna_coordinates_az_el_r is not provided.
                xaxis_info : str
                    The xaxis type. Can be 'Doppler_velocity',
                    'Doppler_frequency' or 'pulse_number'
                vmin : float or None
                    Minimum of the color scale
                vmax : float or None
                    Maximum of the color scale
                plot_type : str
                    Can be 'final' or 'temporal'. If final the data is only
                    plotted at the end of the processing

    Parameters
    ----------
    dataset : spectra
        spectra object

    prdcfg : dictionary of dictionaries
        product configuration dictionary of dictionaries

    Returns
    -------
    None or name of generated files

    """
    dssavedir = prdcfg["dsname"]
    if "dssavename" in prdcfg:
        dssavedir = prdcfg["dssavename"]

    if prdcfg["type"] == "RANGE_DOPPLER":
        field_name = get_fieldname_pyart(prdcfg["voltype"])
        if field_name not in dataset["radar_out"].fields:
            warn(
                " Field type "
                + field_name
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None

        # user defined values
        azi = prdcfg.get("azi", None)
        ele = prdcfg.get("ele", None)
        azi_tol = prdcfg.get("azi_tol", 1.0)
        ele_tol = prdcfg.get("ele_tol", 1.0)

        if azi is None or ele is None:
            ind_ray = prdcfg.get("ind_ray", 0)
            azi = dataset["radar_out"].azimuth["data"][ind_ray]
            ele = dataset["radar_out"].elevation["data"][ind_ray]
        else:
            ind_ray = find_ray_index(
                dataset["radar_out"].elevation["data"],
                dataset["radar_out"].azimuth["data"],
                ele,
                azi,
                ele_tol=ele_tol,
                azi_tol=azi_tol,
            )

        if ind_ray is None:
            warn("Ray azi=" + str(azi) + ", ele=" + str(ele) + " out of radar coverage")
            return None

        gateinfo = "az" + "{:.1f}".format(azi) + "el" + "{:.1f}".format(ele)

        xaxis_info = prdcfg.get("xaxis_info", "Doppler_velocity")
        vmin = prdcfg.get("vmin", None)
        vmax = prdcfg.get("vmax", None)

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=prdcfg["timeinfo"],
        )

        fname_list = make_filename(
            "range_Doppler",
            prdcfg["dstype"],
            prdcfg["voltype"],
            prdcfg["imgformat"],
            prdcfginfo=gateinfo,
            timeinfo=prdcfg["timeinfo"],
            runinfo=prdcfg["runinfo"],
        )

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        if dataset["radar_out"].ngates == 1:
            plot_Doppler(
                dataset["radar_out"],
                field_name,
                ind_ray,
                0,
                prdcfg,
                fname_list,
                xaxis_info=xaxis_info,
                vmin=vmin,
                vmax=vmax,
            )
        else:
            plot_range_Doppler(
                dataset["radar_out"],
                field_name,
                ind_ray,
                prdcfg,
                fname_list,
                xaxis_info=xaxis_info,
                vmin=vmin,
                vmax=vmax,
            )

        print("----- save to " + " ".join(fname_list))

        return fname_list

    if prdcfg["type"] == "ANGLE_DOPPLER":
        field_name = get_fieldname_pyart(prdcfg["voltype"])
        if field_name not in dataset["radar_out"].fields:
            warn(
                " Field type "
                + field_name
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None

        # user defined values
        along_azi = prdcfg.get("along_azi", True)
        ang = prdcfg.get("ang", 0)
        rng = prdcfg.get("rng", 0)
        ang_tol = prdcfg.get("ang_tol", 1.0)
        rng_tol = prdcfg.get("rng_tol", 50.0)

        ind_rng = find_rng_index(
            dataset["radar_out"].range["data"], rng, rng_tol=rng_tol
        )

        if ind_rng is None:
            warn("No data at rng=" + str(rng))
            return None

        if along_azi:
            ind_rays = np.where(
                np.logical_and(
                    dataset["radar_out"].elevation["data"] <= ang + ang_tol,
                    dataset["radar_out"].elevation["data"] >= ang - ang_tol,
                )
            )[0]
        else:
            ind_rays = np.where(
                np.logical_and(
                    dataset["radar_out"].azimuth["data"] <= ang + ang_tol,
                    dataset["radar_out"].azimuth["data"] >= ang - ang_tol,
                )
            )[0]

        if ind_rays.size == 0:
            warn("No data for angle " + str(ang))
            return None

        # sort angles
        if along_azi:
            ang_selected = dataset["radar_out"].azimuth["data"][ind_rays]

        else:
            ang_selected = dataset["radar_out"].elevation["data"][ind_rays]
        ind_rays = ind_rays[np.argsort(ang_selected)]

        if along_azi:
            gateinfo = "azi" + "{:.1f}".format(ang) + "rng" + "{:.1f}".format(rng)
        else:
            gateinfo = "ele" + "{:.1f}".format(ang) + "rng" + "{:.1f}".format(rng)

        xaxis_info = prdcfg.get("xaxis_info", "Doppler_velocity")
        vmin = prdcfg.get("vmin", None)
        vmax = prdcfg.get("vmax", None)

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=prdcfg["timeinfo"],
        )

        fname_list = make_filename(
            "range_Doppler",
            prdcfg["dstype"],
            prdcfg["voltype"],
            prdcfg["imgformat"],
            prdcfginfo=gateinfo,
            timeinfo=prdcfg["timeinfo"],
            runinfo=prdcfg["runinfo"],
        )

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        if ind_rays.size == 1:
            plot_Doppler(
                dataset["radar_out"],
                field_name,
                ind_rays,
                ind_rng,
                prdcfg,
                fname_list,
                xaxis_info=xaxis_info,
                vmin=vmin,
                vmax=vmax,
            )
        else:
            plot_angle_Doppler(
                dataset["radar_out"],
                field_name,
                ang,
                ind_rays,
                ind_rng,
                prdcfg,
                fname_list,
                xaxis_info=xaxis_info,
                along_azi=along_azi,
                vmin=vmin,
                vmax=vmax,
            )

        print("----- save to " + " ".join(fname_list))

        return fname_list

    if prdcfg["type"] == "TIME_DOPPLER":
        field_name = get_fieldname_pyart(prdcfg["voltype"])
        if field_name not in dataset["radar_out"].fields:
            warn(
                " Field type "
                + field_name
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None

        # user defined values
        xaxis_info = prdcfg.get("xaxis_info", "Doppler_velocity")
        vmin = prdcfg.get("vmin", None)
        vmax = prdcfg.get("vmax", None)
        xmin = prdcfg.get("xmin", None)
        xmax = prdcfg.get("xmax", None)
        ymin = prdcfg.get("ymin", None)
        ymax = prdcfg.get("ymax", None)
        plot_type = prdcfg.get("plot_type", "final")

        if plot_type == "final" and not dataset["final"]:
            return None

        if "antenna_coordinates_az_el_r" in dataset:
            az = "{:.1f}".format(dataset["antenna_coordinates_az_el_r"][0])
            el = "{:.1f}".format(dataset["antenna_coordinates_az_el_r"][1])
            r = "{:.1f}".format(dataset["antenna_coordinates_az_el_r"][2])
            gateinfo = "az" + az + "r" + r + "el" + el
        else:
            lon = "{:.3f}".format(dataset["point_coordinates_WGS84_lon_lat_alt"][0])
            lat = "{:.3f}".format(dataset["point_coordinates_WGS84_lon_lat_alt"][1])
            alt = "{:.1f}".format(dataset["point_coordinates_WGS84_lon_lat_alt"][2])
            gateinfo = "lon" + lon + "lat" + lat + "alt" + alt

        time_info = datetime_from_radar(dataset["radar_out"])

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=time_info,
        )

        fname_list = make_filename(
            "time_Doppler",
            prdcfg["dstype"],
            prdcfg["voltype"],
            prdcfg["imgformat"],
            prdcfginfo=gateinfo,
            timeinfo=time_info,
            runinfo=prdcfg["runinfo"],
        )

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        if dataset["radar_out"].nrays == 1:
            plot_Doppler(
                dataset["radar_out"],
                field_name,
                0,
                0,
                prdcfg,
                fname_list,
                xaxis_info=xaxis_info,
                vmin=vmin,
                vmax=vmax,
            )
        else:
            plot_time_Doppler(
                dataset["radar_out"],
                field_name,
                prdcfg,
                fname_list,
                xaxis_info=xaxis_info,
                vmin=vmin,
                vmax=vmax,
                xmin=xmin,
                xmax=xmax,
                ymin=ymin,
                ymax=ymax,
            )

        print("----- save to " + " ".join(fname_list))

        return fname_list

    if prdcfg["type"] == "DOPPLER":
        field_name = get_fieldname_pyart(prdcfg["voltype"])
        if field_name not in dataset["radar_out"].fields:
            warn(
                " Field type "
                + field_name
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None

        # user defined values
        azi = prdcfg.get("azi", None)
        ele = prdcfg.get("ele", None)
        rng = prdcfg.get("rng", None)
        azi_tol = prdcfg.get("azi_tol", 1.0)
        ele_tol = prdcfg.get("ele_tol", 1.0)
        rng_tol = prdcfg.get("rng_tol", 50.0)

        if azi is None or ele is None or rng is None:
            ind_ray = prdcfg.get("ind_ray", 0)
            ind_rng = prdcfg.get("ind_rng", 0)
            azi = dataset["radar_out"].azimuth["data"][ind_ray]
            ele = dataset["radar_out"].elevation["data"][ind_ray]
            rng = dataset["radar_out"].range["data"][ind_rng]
        else:
            ind_ray = find_ray_index(
                dataset["radar_out"].elevation["data"],
                dataset["radar_out"].azimuth["data"],
                ele,
                azi,
                ele_tol=ele_tol,
                azi_tol=azi_tol,
            )
            ind_rng = find_rng_index(
                dataset["radar_out"].range["data"], rng, rng_tol=rng_tol
            )

        if ind_rng is None or ind_ray is None:
            warn(
                "Point azi="
                + str(azi)
                + ", ele="
                + str(ele)
                + ", rng="
                + str(rng)
                + " out of radar coverage"
            )
            return None

        gateinfo = (
            "az"
            + "{:.1f}".format(azi)
            + "el"
            + "{:.1f}".format(ele)
            + "r"
            + "{:.1f}".format(rng)
        )

        xaxis_info = prdcfg.get("xaxis_info", "Doppler_velocity")
        vmin = prdcfg.get("vmin", None)
        vmax = prdcfg.get("vmax", None)

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=prdcfg["timeinfo"],
        )

        fname_list = make_filename(
            "Doppler",
            prdcfg["dstype"],
            prdcfg["voltype"],
            prdcfg["imgformat"],
            prdcfginfo=gateinfo,
            timeinfo=prdcfg["timeinfo"],
            runinfo=prdcfg["runinfo"],
        )

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        plot_Doppler(
            dataset["radar_out"],
            field_name,
            ind_ray,
            ind_rng,
            prdcfg,
            fname_list,
            xaxis_info=xaxis_info,
            vmin=vmin,
            vmax=vmax,
        )

        print("----- save to " + " ".join(fname_list))

        return fname_list

    if prdcfg["type"] == "COMPLEX_RANGE_DOPPLER":
        field_name = get_fieldname_pyart(prdcfg["voltype"])
        if field_name not in dataset["radar_out"].fields:
            warn(
                " Field type "
                + field_name
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None

        # user defined values
        azi = prdcfg.get("azi", None)
        ele = prdcfg.get("ele", None)
        azi_tol = prdcfg.get("azi_tol", 1.0)
        ele_tol = prdcfg.get("ele_tol", 1.0)

        if azi is None or ele is None:
            ind_ray = prdcfg.get("ind_ray", 0)
            azi = dataset["radar_out"].azimuth["data"][ind_ray]
            ele = dataset["radar_out"].elevation["data"][ind_ray]
        else:
            ind_ray = find_ray_index(
                dataset["radar_out"].elevation["data"],
                dataset["radar_out"].azimuth["data"],
                ele,
                azi,
                ele_tol=ele_tol,
                azi_tol=azi_tol,
            )

        if ind_ray is None:
            warn("Ray azi=" + str(azi) + ", ele=" + str(ele) + " out of radar coverage")
            return None

        gateinfo = "az" + "{:.1f}".format(azi) + "el" + "{:.1f}".format(ele)

        xaxis_info = prdcfg.get("xaxis_info", "Doppler_velocity")
        vmin = prdcfg.get("vmin", None)
        vmax = prdcfg.get("vmax", None)

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=prdcfg["timeinfo"],
        )

        fname_list = make_filename(
            "c_range_Doppler",
            prdcfg["dstype"],
            prdcfg["voltype"],
            prdcfg["imgformat"],
            prdcfginfo=gateinfo,
            timeinfo=prdcfg["timeinfo"],
            runinfo=prdcfg["runinfo"],
        )

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        if dataset["radar_out"].ngates == 1:
            plot_complex_Doppler(
                dataset["radar_out"],
                field_name,
                ind_ray,
                0,
                prdcfg,
                fname_list,
                xaxis_info=xaxis_info,
                vmin=vmin,
                vmax=vmax,
            )
        else:
            plot_complex_range_Doppler(
                dataset["radar_out"],
                field_name,
                ind_ray,
                prdcfg,
                fname_list,
                xaxis_info=xaxis_info,
                vmin=vmin,
                vmax=vmax,
            )

        print("----- save to " + " ".join(fname_list))

        return fname_list

    if prdcfg["type"] == "COMPLEX_ANGLE_DOPPLER":
        field_name = get_fieldname_pyart(prdcfg["voltype"])
        if field_name not in dataset["radar_out"].fields:
            warn(
                " Field type "
                + field_name
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None

        # user defined values
        along_azi = prdcfg.get("along_azi", True)
        ang = prdcfg.get("ang", 0)
        rng = prdcfg.get("rng", 0)
        ang_tol = prdcfg.get("ang_tol", 1.0)
        rng_tol = prdcfg.get("rng_tol", 50.0)

        ind_rng = find_rng_index(
            dataset["radar_out"].range["data"], rng, rng_tol=rng_tol
        )

        if ind_rng is None:
            warn("No data at rng=" + str(rng))
            return None

        if along_azi:
            ind_rays = np.where(
                np.logical_and(
                    dataset["radar_out"].elevation["data"] <= ang + ang_tol,
                    dataset["radar_out"].elevation["data"] >= ang - ang_tol,
                )
            )[0]
        else:
            ind_rays = np.where(
                np.logical_and(
                    dataset["radar_out"].azimuth["data"] <= ang + ang_tol,
                    dataset["radar_out"].azimuth["data"] >= ang - ang_tol,
                )
            )[0]

        if ind_rays.size == 0:
            warn("No data for angle " + str(ang))
            return None

        # sort angles
        if along_azi:
            ang_selected = dataset["radar_out"].azimuth["data"][ind_rays]

        else:
            ang_selected = dataset["radar_out"].elevation["data"][ind_rays]
        ind_rays = ind_rays[np.argsort(ang_selected)]

        if along_azi:
            gateinfo = "azi" + "{:.1f}".format(ang) + "rng" + "{:.1f}".format(rng)
        else:
            gateinfo = "ele" + "{:.1f}".format(ang) + "rng" + "{:.1f}".format(rng)

        xaxis_info = prdcfg.get("xaxis_info", "Doppler_velocity")
        vmin = prdcfg.get("vmin", None)
        vmax = prdcfg.get("vmax", None)

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=prdcfg["timeinfo"],
        )

        fname_list = make_filename(
            "range_Doppler",
            prdcfg["dstype"],
            prdcfg["voltype"],
            prdcfg["imgformat"],
            prdcfginfo=gateinfo,
            timeinfo=prdcfg["timeinfo"],
            runinfo=prdcfg["runinfo"],
        )

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        if ind_rays.size == 1:
            plot_complex_Doppler(
                dataset["radar_out"],
                field_name,
                ind_rays,
                ind_rng,
                prdcfg,
                fname_list,
                xaxis_info=xaxis_info,
                vmin=vmin,
                vmax=vmax,
            )
        else:
            plot_complex_angle_Doppler(
                dataset["radar_out"],
                field_name,
                ang,
                ind_rays,
                ind_rng,
                prdcfg,
                fname_list,
                xaxis_info=xaxis_info,
                along_azi=along_azi,
                vmin=vmin,
                vmax=vmax,
            )

        print("----- save to " + " ".join(fname_list))

        return fname_list

    if prdcfg["type"] == "COMPLEX_TIME_DOPPLER":
        field_name = get_fieldname_pyart(prdcfg["voltype"])
        if field_name not in dataset["radar_out"].fields:
            warn(
                " Field type "
                + field_name
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None

        # user defined values
        xaxis_info = prdcfg.get("xaxis_info", "Doppler_velocity")
        vmin = prdcfg.get("vmin", None)
        vmax = prdcfg.get("vmax", None)
        plot_type = prdcfg.get("plot_type", "final")

        if plot_type == "final" and not dataset["final"]:
            return None

        if "antenna_coordinates_az_el_r" in dataset:
            az = "{:.1f}".format(dataset["antenna_coordinates_az_el_r"][0])
            el = "{:.1f}".format(dataset["antenna_coordinates_az_el_r"][1])
            r = "{:.1f}".format(dataset["antenna_coordinates_az_el_r"][2])
            gateinfo = "az" + az + "r" + r + "el" + el
        else:
            lon = "{:.3f}".format(dataset["point_coordinates_WGS84_lon_lat_alt"][0])
            lat = "{:.3f}".format(dataset["point_coordinates_WGS84_lon_lat_alt"][1])
            alt = "{:.1f}".format(dataset["point_coordinates_WGS84_lon_lat_alt"][2])
            gateinfo = "lon" + lon + "lat" + lat + "alt" + alt

        time_info = datetime_from_radar(dataset["radar_out"])

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=time_info,
        )

        fname_list = make_filename(
            "c_time_Doppler",
            prdcfg["dstype"],
            prdcfg["voltype"],
            prdcfg["imgformat"],
            prdcfginfo=gateinfo,
            timeinfo=time_info,
            runinfo=prdcfg["runinfo"],
        )

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        if dataset["radar_out"].nrays == 1:
            plot_complex_Doppler(
                dataset["radar_out"],
                field_name,
                0,
                0,
                prdcfg,
                fname_list,
                xaxis_info=xaxis_info,
                vmin=vmin,
                vmax=vmax,
            )
        else:
            plot_complex_time_Doppler(
                dataset["radar_out"],
                field_name,
                prdcfg,
                fname_list,
                xaxis_info=xaxis_info,
                vmin=vmin,
                vmax=vmax,
            )

        print("----- save to " + " ".join(fname_list))

        return fname_list

    if prdcfg["type"] == "COMPLEX_DOPPLER":
        field_name = get_fieldname_pyart(prdcfg["voltype"])
        if field_name not in dataset["radar_out"].fields:
            warn(
                " Field type "
                + field_name
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None

        # user defined values
        azi = prdcfg.get("azi", None)
        ele = prdcfg.get("ele", None)
        rng = prdcfg.get("rng", None)
        azi_tol = prdcfg.get("azi_tol", 1.0)
        ele_tol = prdcfg.get("ele_tol", 1.0)
        rng_tol = prdcfg.get("rng_tol", 50.0)

        if azi is None or ele is None or rng is None:
            ind_ray = prdcfg.get("ind_ray", 0)
            ind_rng = prdcfg.get("ind_rng", 0)
            azi = dataset["radar_out"].azimuth["data"][ind_ray]
            ele = dataset["radar_out"].elevation["data"][ind_ray]
            rng = dataset["radar_out"].range["data"][ind_rng]
        else:
            ind_ray = find_ray_index(
                dataset["radar_out"].elevation["data"],
                dataset["radar_out"].azimuth["data"],
                ele,
                azi,
                ele_tol=ele_tol,
                azi_tol=azi_tol,
            )
            ind_rng = find_rng_index(
                dataset["radar_out"].range["data"], rng, rng_tol=rng_tol
            )

        if ind_rng is None or ind_ray is None:
            warn(
                "Point azi="
                + str(azi)
                + ", ele="
                + str(ele)
                + ", rng="
                + str(rng)
                + " out of radar coverage"
            )
            return None

        gateinfo = (
            "az"
            + "{:.1f}".format(azi)
            + "el"
            + "{:.1f}".format(ele)
            + "r"
            + "{:.1f}".format(rng)
        )

        xaxis_info = prdcfg.get("xaxis_info", "Doppler_velocity")
        vmin = prdcfg.get("vmin", None)
        vmax = prdcfg.get("vmax", None)

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=prdcfg["timeinfo"],
        )

        fname_list = make_filename(
            "c_Doppler",
            prdcfg["dstype"],
            prdcfg["voltype"],
            prdcfg["imgformat"],
            prdcfginfo=gateinfo,
            timeinfo=prdcfg["timeinfo"],
            runinfo=prdcfg["runinfo"],
        )

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        plot_complex_Doppler(
            dataset["radar_out"],
            field_name,
            ind_ray,
            ind_rng,
            prdcfg,
            fname_list,
            xaxis_info=xaxis_info,
            vmin=vmin,
            vmax=vmax,
        )

        print("----- save to " + " ".join(fname_list))

        return fname_list

    if prdcfg["type"] == "AMPLITUDE_PHASE_DOPPLER":
        field_name = get_fieldname_pyart(prdcfg["voltype"])
        if field_name not in dataset["radar_out"].fields:
            warn(
                " Field type "
                + field_name
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None

        # user defined values
        azi = prdcfg.get("azi", None)
        ele = prdcfg.get("ele", None)
        rng = prdcfg.get("rng", None)
        azi_tol = prdcfg.get("azi_tol", 1.0)
        ele_tol = prdcfg.get("ele_tol", 1.0)
        rng_tol = prdcfg.get("rng_tol", 50.0)

        if azi is None or ele is None or rng is None:
            ind_ray = prdcfg.get("ind_ray", 0)
            ind_rng = prdcfg.get("ind_rng", 0)
            azi = dataset["radar_out"].azimuth["data"][ind_ray]
            ele = dataset["radar_out"].elevation["data"][ind_ray]
            rng = dataset["radar_out"].range["data"][ind_rng]
        else:
            ind_ray = find_ray_index(
                dataset["radar_out"].elevation["data"],
                dataset["radar_out"].azimuth["data"],
                ele,
                azi,
                ele_tol=ele_tol,
                azi_tol=azi_tol,
            )
            ind_rng = find_rng_index(
                dataset["radar_out"].range["data"], rng, rng_tol=rng_tol
            )

        if ind_rng is None or ind_ray is None:
            warn(
                "Point azi="
                + str(azi)
                + ", ele="
                + str(ele)
                + ", rng="
                + str(rng)
                + " out of radar coverage"
            )
            return None

        gateinfo = (
            "az"
            + "{:.1f}".format(azi)
            + "el"
            + "{:.1f}".format(ele)
            + "r"
            + "{:.1f}".format(rng)
        )

        xaxis_info = prdcfg.get("xaxis_info", "Doppler_velocity")
        ampli_vmin = prdcfg.get("ampli_vmin", None)
        ampli_vmax = prdcfg.get("ampli_vmax", None)
        phase_vmin = prdcfg.get("phase_vmin", None)
        phase_vmax = prdcfg.get("phase_vmax", None)

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=prdcfg["timeinfo"],
        )

        fname_list = make_filename(
            "ap_Doppler",
            prdcfg["dstype"],
            prdcfg["voltype"],
            prdcfg["imgformat"],
            prdcfginfo=gateinfo,
            timeinfo=prdcfg["timeinfo"],
            runinfo=prdcfg["runinfo"],
        )

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        plot_amp_phase_Doppler(
            dataset["radar_out"],
            field_name,
            ind_ray,
            ind_rng,
            prdcfg,
            fname_list,
            xaxis_info=xaxis_info,
            ampli_vmin=ampli_vmin,
            ampli_vmax=ampli_vmax,
            phase_vmin=phase_vmin,
            phase_vmax=phase_vmax,
        )

        print("----- save to " + " ".join(fname_list))

        return fname_list

    if prdcfg["type"] == "AMPLITUDE_PHASE_RANGE_DOPPLER":
        field_name = get_fieldname_pyart(prdcfg["voltype"])
        if field_name not in dataset["radar_out"].fields:
            warn(
                " Field type "
                + field_name
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None

        # user defined values
        azi = prdcfg.get("azi", None)
        ele = prdcfg.get("ele", None)
        azi_tol = prdcfg.get("azi_tol", 1.0)
        ele_tol = prdcfg.get("ele_tol", 1.0)

        if azi is None or ele is None:
            ind_ray = prdcfg.get("ind_ray", 0)
            azi = dataset["radar_out"].azimuth["data"][ind_ray]
            ele = dataset["radar_out"].elevation["data"][ind_ray]
        else:
            ind_ray = find_ray_index(
                dataset["radar_out"].elevation["data"],
                dataset["radar_out"].azimuth["data"],
                ele,
                azi,
                ele_tol=ele_tol,
                azi_tol=azi_tol,
            )

        if ind_ray is None:
            warn("Ray azi=" + str(azi) + ", ele=" + str(ele) + " out of radar coverage")
            return None

        gateinfo = "az" + "{:.1f}".format(azi) + "el" + "{:.1f}".format(ele)

        xaxis_info = prdcfg.get("xaxis_info", "Doppler_velocity")
        ampli_vmin = prdcfg.get("ampli_vmin", None)
        ampli_vmax = prdcfg.get("ampli_vmax", None)
        phase_vmin = prdcfg.get("phase_vmin", None)
        phase_vmax = prdcfg.get("phase_vmax", None)

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=prdcfg["timeinfo"],
        )

        fname_list = make_filename(
            "ap_range_Doppler",
            prdcfg["dstype"],
            prdcfg["voltype"],
            prdcfg["imgformat"],
            prdcfginfo=gateinfo,
            timeinfo=prdcfg["timeinfo"],
            runinfo=prdcfg["runinfo"],
        )

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        if dataset["radar_out"].ngates == 1:
            plot_amp_phase_Doppler(
                dataset["radar_out"],
                field_name,
                ind_ray,
                0,
                prdcfg,
                fname_list,
                xaxis_info=xaxis_info,
                ampli_vmin=ampli_vmin,
                ampli_vmax=ampli_vmax,
                phase_vmin=phase_vmin,
                phase_vmax=phase_vmax,
            )
        else:
            plot_amp_phase_range_Doppler(
                dataset["radar_out"],
                field_name,
                ind_ray,
                prdcfg,
                fname_list,
                xaxis_info=xaxis_info,
                ampli_vmin=ampli_vmin,
                ampli_vmax=ampli_vmax,
                phase_vmin=phase_vmin,
                phase_vmax=phase_vmax,
            )

        print("----- save to " + " ".join(fname_list))

        return fname_list

    if prdcfg["type"] == "AMPLITUDE_PHASE_ANGLE_DOPPLER":
        field_name = get_fieldname_pyart(prdcfg["voltype"])
        if field_name not in dataset["radar_out"].fields:
            warn(
                " Field type "
                + field_name
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None

        # user defined values
        along_azi = prdcfg.get("along_azi", True)
        ang = prdcfg.get("ang", 0)
        rng = prdcfg.get("rng", 0)
        ang_tol = prdcfg.get("ang_tol", 1.0)
        rng_tol = prdcfg.get("rng_tol", 50.0)

        ind_rng = find_rng_index(
            dataset["radar_out"].range["data"], rng, rng_tol=rng_tol
        )

        if ind_rng is None:
            warn("No data at rng=" + str(rng))
            return None

        if along_azi:
            ind_rays = np.where(
                np.logical_and(
                    dataset["radar_out"].elevation["data"] <= ang + ang_tol,
                    dataset["radar_out"].elevation["data"] >= ang - ang_tol,
                )
            )[0]
        else:
            ind_rays = np.where(
                np.logical_and(
                    dataset["radar_out"].azimuth["data"] <= ang + ang_tol,
                    dataset["radar_out"].azimuth["data"] >= ang - ang_tol,
                )
            )[0]

        if ind_rays.size == 0:
            warn("No data for angle " + str(ang))
            return None

        # sort angles
        if along_azi:
            ang_selected = dataset["radar_out"].azimuth["data"][ind_rays]

        else:
            ang_selected = dataset["radar_out"].elevation["data"][ind_rays]
        ind_rays = ind_rays[np.argsort(ang_selected)]

        if along_azi:
            gateinfo = "azi" + "{:.1f}".format(ang) + "rng" + "{:.1f}".format(rng)
        else:
            gateinfo = "ele" + "{:.1f}".format(ang) + "rng" + "{:.1f}".format(rng)

        xaxis_info = prdcfg.get("xaxis_info", "Doppler_velocity")
        ampli_vmin = prdcfg.get("ampli_vmin", None)
        ampli_vmax = prdcfg.get("ampli_vmax", None)
        phase_vmin = prdcfg.get("phase_vmin", None)
        phase_vmax = prdcfg.get("phase_vmax", None)

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=prdcfg["timeinfo"],
        )

        fname_list = make_filename(
            "range_Doppler",
            prdcfg["dstype"],
            prdcfg["voltype"],
            prdcfg["imgformat"],
            prdcfginfo=gateinfo,
            timeinfo=prdcfg["timeinfo"],
            runinfo=prdcfg["runinfo"],
        )

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        if ind_rays.size == 1:
            plot_amp_phase_Doppler(
                dataset["radar_out"],
                field_name,
                ind_rays,
                ind_rng,
                prdcfg,
                fname_list,
                xaxis_info=xaxis_info,
                ampli_vmin=ampli_vmin,
                ampli_vmax=ampli_vmax,
                phase_vmin=phase_vmin,
                phase_vmax=phase_vmax,
            )
        else:
            plot_amp_phase_angle_Doppler(
                dataset["radar_out"],
                field_name,
                ang,
                ind_rays,
                ind_rng,
                prdcfg,
                fname_list,
                xaxis_info=xaxis_info,
                along_azi=along_azi,
                ampli_vmin=ampli_vmin,
                ampli_vmax=ampli_vmax,
                phase_vmin=phase_vmin,
                phase_vmax=phase_vmax,
            )

        print("----- save to " + " ".join(fname_list))

        return fname_list

    if prdcfg["type"] == "AMPLITUDE_PHASE_TIME_DOPPLER":
        field_name = get_fieldname_pyart(prdcfg["voltype"])
        if field_name not in dataset["radar_out"].fields:
            warn(
                " Field type "
                + field_name
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None

        # user defined values
        xaxis_info = prdcfg.get("xaxis_info", "Doppler_velocity")
        ampli_vmin = prdcfg.get("ampli_vmin", None)
        ampli_vmax = prdcfg.get("ampli_vmax", None)
        phase_vmin = prdcfg.get("phase_vmin", None)
        phase_vmax = prdcfg.get("phase_vmax", None)
        plot_type = prdcfg.get("plot_type", "final")

        if plot_type == "final" and not dataset["final"]:
            return None

        if "antenna_coordinates_az_el_r" in dataset:
            az = "{:.1f}".format(dataset["antenna_coordinates_az_el_r"][0])
            el = "{:.1f}".format(dataset["antenna_coordinates_az_el_r"][1])
            r = "{:.1f}".format(dataset["antenna_coordinates_az_el_r"][2])
            gateinfo = "az" + az + "r" + r + "el" + el
        else:
            lon = "{:.3f}".format(dataset["point_coordinates_WGS84_lon_lat_alt"][0])
            lat = "{:.3f}".format(dataset["point_coordinates_WGS84_lon_lat_alt"][1])
            alt = "{:.1f}".format(dataset["point_coordinates_WGS84_lon_lat_alt"][2])
            gateinfo = "lon" + lon + "lat" + lat + "alt" + alt

        time_info = datetime_from_radar(dataset["radar_out"])

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=time_info,
        )

        fname_list = make_filename(
            "ap_time_Doppler",
            prdcfg["dstype"],
            prdcfg["voltype"],
            prdcfg["imgformat"],
            prdcfginfo=gateinfo,
            timeinfo=time_info,
            runinfo=prdcfg["runinfo"],
        )

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        if dataset["radar_out"].nrays == 1:
            plot_amp_phase_Doppler(
                dataset["radar_out"],
                field_name,
                0,
                0,
                prdcfg,
                fname_list,
                xaxis_info=xaxis_info,
                ampli_vmin=ampli_vmin,
                ampli_vmax=ampli_vmax,
                phase_vmin=phase_vmin,
                phase_vmax=phase_vmax,
            )
        else:
            plot_amp_phase_time_Doppler(
                dataset["radar_out"],
                field_name,
                prdcfg,
                fname_list,
                xaxis_info=xaxis_info,
                ampli_vmin=ampli_vmin,
                ampli_vmax=ampli_vmax,
                phase_vmin=phase_vmin,
                phase_vmax=phase_vmax,
            )

        print("----- save to " + " ".join(fname_list))

        return fname_list

    if prdcfg["type"] == "SAVEVOL":
        field_name = get_fieldname_pyart(prdcfg["voltype"])
        if field_name not in dataset["radar_out"].fields:
            warn(
                " Field type "
                + field_name
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None

        file_type = prdcfg.get("file_type", "nc")
        prdcfg.get("physical", True)

        new_dataset = deepcopy(dataset["radar_out"])
        new_dataset.fields = dict()
        new_dataset.add_field(field_name, dataset["radar_out"].fields[field_name])

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=prdcfg["timeinfo"],
        )

        fname = make_filename(
            "savevol",
            prdcfg["dstype"],
            prdcfg["voltype"],
            [file_type],
            timeinfo=prdcfg["timeinfo"],
            runinfo=prdcfg["runinfo"],
        )[0]

        fname = savedir + fname

        pyart.aux_io.write_spectra(fname, new_dataset)

        print("saved file: " + fname)

        return fname

    if prdcfg["type"] == "SAVEALL":
        file_type = prdcfg.get("file_type", "nc")
        datatypes = prdcfg.get("datatypes", None)
        prdcfg.get("physical", True)

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=prdcfg["timeinfo"],
        )

        fname = make_filename(
            "savevol",
            prdcfg["dstype"],
            "all_fields",
            [file_type],
            timeinfo=prdcfg["timeinfo"],
            runinfo=prdcfg["runinfo"],
        )[0]

        fname = savedir + fname

        field_names = None
        if datatypes is not None:
            field_names = []
            for datatype in datatypes:
                field_names.append(get_fieldname_pyart(datatype))

        if field_names is not None:
            radar_aux = deepcopy(dataset["radar_out"])
            radar_aux.fields = dict()
            for field_name in field_names:
                if field_name not in dataset["radar_out"].fields:
                    warn(field_name + " not in radar object")
                else:
                    radar_aux.add_field(
                        field_name, dataset["radar_out"].fields[field_name]
                    )
        else:
            radar_aux = dataset["radar_out"]
        pyart.aux_io.write_spectra(
            fname,
            radar_aux,
        )

        print("saved file: " + fname)

        return fname

    warn(" Unsupported product type: " + prdcfg["type"])
    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
================================================
common_colocated_gates
================================================

This program reads colocated gates files from two radars
and creates a new file with the gates that are common to
both radars

"""

# Author: fvj
# License: BSD 3 clause

import datetime
import atexit
import numpy as np
import pandas as pd

from pyrad.io import read_colocated_gates, write_colocated_gates

print(__doc__)


def main():
    """ """

    file_path = "/srn/analysis/pyrad_products/rad4alp_intercomp/colocated_gates/"
    rad1_vec = ["A", "A", "A", "A", "D", "D", "D", "L", "L", "P"]
    rad2_vec = ["D", "L", "P", "W", "L", "P", "W", "P", "W", "W"]

    print(
        "====== common colocated gates started: %s"
        % datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
    )
    atexit.register(_print_end_msg, "====== common colocated gates finished: ")

    for i, rad1 in enumerate(rad1_vec):
        rad2 = rad2_vec[i]

        print("Radars: " + rad1 + " " + rad2)
        fname1 = (
            file_path
            + "PL"
            + rad1
            + "_"
            + "PL"
            + rad2
            + "/info_COLOCATED_GATES_PL"
            + rad1
            + "_PL"
            + rad2
            + ".csv"
        )
        fname2 = (
            file_path
            + "PL"
            + rad2
            + "_"
            + "PL"
            + rad1
            + "/info_COLOCATED_GATES_PL"
            + rad2
            + "_PL"
            + rad1
            + ".csv"
        )

        (
            rad1_ray_ind,
            rad1_rng_ind,
            rad1_ele,
            rad1_azi,
            rad1_rng,
            rad2_ray_ind,
            rad2_rng_ind,
            rad2_ele,
            rad2_azi,
            rad2_rng,
        ) = read_colocated_gates(fname1)

        print("Number of gates rad1-rad2 ", np.shape(rad1_ray_ind))

        (
            rad2_ray_ind_aux,
            rad2_rng_ind_aux,
            rad2_ele_aux,
            rad2_azi_aux,
            rad2_rng_aux,
            rad1_ray_ind_aux,
            rad1_rng_ind_aux,
            rad1_ele_aux,
            rad1_azi_aux,
            rad1_rng_aux,
        ) = read_colocated_gates(fname2)

        print("Number of gates rad2-rad1 ", np.shape(rad2_ray_ind_aux))

        # make a pool of data
        rad1_ray_ind = np.ma.concatenate((rad1_ray_ind, rad1_ray_ind_aux))
        rad1_rng_ind = np.ma.concatenate((rad1_rng_ind, rad1_rng_ind_aux))
        rad1_ele = np.ma.concatenate((rad1_ele, rad1_ele_aux))
        rad1_azi = np.ma.concatenate((rad1_azi, rad1_azi_aux))
        rad1_rng = np.ma.concatenate((rad1_rng, rad1_rng_aux))
        rad2_ray_ind = np.ma.concatenate((rad2_ray_ind, rad2_ray_ind_aux))
        rad2_rng_ind = np.ma.concatenate((rad2_rng_ind, rad2_rng_ind_aux))
        rad2_ele = np.ma.concatenate((rad2_ele, rad2_ele_aux))
        rad2_azi = np.ma.concatenate((rad2_azi, rad2_azi_aux))
        rad2_rng = np.ma.concatenate((rad2_rng, rad2_rng_aux))

        print("Total number of gates ", np.shape(rad1_ray_ind))

        # create dictionary and put it in pandas framework
        coloc_dict = {
            "rad1_ray_ind": rad1_ray_ind,
            "rad1_rng_ind": rad1_rng_ind,
            "rad1_ele": rad1_ele,
            "rad1_azi": rad1_azi,
            "rad1_rng": rad1_rng,
            "rad2_ray_ind": rad2_ray_ind,
            "rad2_rng_ind": rad2_rng_ind,
            "rad2_ele": rad2_ele,
            "rad2_azi": rad2_azi,
            "rad2_rng": rad2_rng,
        }
        df = pd.DataFrame(data=coloc_dict)

        # keep only duplicated data
        df_common = df[df.duplicated(keep=False)].drop_duplicates()
        common_dict = df_common.to_dict(orient="list")

        print("Number of common gates", df_common.shape)
        print(
            "rad1 elev min/max",
            np.min(common_dict["rad1_ele"]),
            np.max(common_dict["rad1_ele"]),
        )
        print(
            "rad2 elev min/max",
            np.min(common_dict["rad2_ele"]),
            np.max(common_dict["rad2_ele"]),
        )

        # write resultant output
        fname1_out = (
            file_path
            + "PL"
            + rad1
            + "_"
            + "PL"
            + rad2
            + "/info_common_COLOCATED_GATES_PL"
            + rad1
            + "_PL"
            + rad2
            + ".csv"
        )
        fname2_out = (
            file_path
            + "PL"
            + rad2
            + "_"
            + "PL"
            + rad1
            + "/info_common_COLOCATED_GATES_PL"
            + rad2
            + "_PL"
            + rad1
            + ".csv"
        )

        rad1_dict = {
            "rad1_ray_ind": np.asarray(common_dict["rad1_ray_ind"]),
            "rad1_rng_ind": np.asarray(common_dict["rad1_rng_ind"]),
            "rad1_ele": np.asarray(common_dict["rad1_ele"]),
            "rad1_azi": np.asarray(common_dict["rad1_azi"]),
            "rad1_rng": np.asarray(common_dict["rad1_rng"]),
            "rad2_ray_ind": np.asarray(common_dict["rad2_ray_ind"]),
            "rad2_rng_ind": np.asarray(common_dict["rad2_rng_ind"]),
            "rad2_ele": np.asarray(common_dict["rad2_ele"]),
            "rad2_azi": np.asarray(common_dict["rad2_azi"]),
            "rad2_rng": np.asarray(common_dict["rad2_rng"]),
        }

        rad2_dict = {
            "rad1_ray_ind": np.asarray(common_dict["rad2_ray_ind"]),
            "rad1_rng_ind": np.asarray(common_dict["rad2_rng_ind"]),
            "rad1_ele": np.asarray(common_dict["rad2_ele"]),
            "rad1_azi": np.asarray(common_dict["rad2_azi"]),
            "rad1_rng": np.asarray(common_dict["rad2_rng"]),
            "rad2_ray_ind": np.asarray(common_dict["rad1_ray_ind"]),
            "rad2_rng_ind": np.asarray(common_dict["rad1_rng_ind"]),
            "rad2_ele": np.asarray(common_dict["rad1_ele"]),
            "rad2_azi": np.asarray(common_dict["rad1_azi"]),
            "rad2_rng": np.asarray(common_dict["rad1_rng"]),
        }

        write_colocated_gates(rad1_dict, fname1_out)
        write_colocated_gates(rad2_dict, fname2_out)


def _print_end_msg(text):
    """
    prints end message

    Parameters
    ----------
    text : str
        the text to be printed

    Returns
    -------
    Nothing

    """
    print(
        text
        + datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
    )


# ---------------------------------------------------------
# Start main:
# ---------------------------------------------------------
if __name__ == "__main__":
    main()



"""
pyrad.proc.process_traj
=============================

Trajectory functions. Functions to pass trajectory dataset data to
the product generation functions.

.. autosummary::
    :toctree: generated/

    process_trajectory
    process_traj_trt
    process_traj_trt_contour
    process_traj_lightning
    process_traj_atplane
    process_traj_antenna_pattern
    _get_ts_values_antenna_pattern
    _get_contour_trt
    _get_gates
    _get_gates_trt
    _get_gates_antenna_pattern
    _get_closest_bin
    _sample_out_of_sector
    TargetRadar
"""

from warnings import warn
import gc
from copy import deepcopy

import numpy as np
from netCDF4 import num2date, date2num

import pyart
from pyart.config import get_metadata
from pyart.core import Radar

try:
    from pyart.util import get_target_elevations, cross_section_rhi
except ImportError:
    warn("ARM version of Py-ART detected, you will not be able to use some products")
    warn("Please use Py-ART MCH instead (https://github.com/MeteoSwiss/pyart)")


from ..io.io_aux import get_datatype_fields, get_fieldname_pyart
from ..io.io_aux import get_field_unit, get_field_name
from ..io.timeseries import TimeSeries
from ..io.read_data_other import read_antenna_pattern

from ..util.stat_utils import quantiles_weighted
from ..util.radar_utils import belongs_roi_indices, find_nearest_gate


def process_trajectory(procstatus, dscfg, radar_list=None, trajectory=None):
    """
    Return trajectory

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types, can be any datatype supported by pyrad
            and available in the radar data
    radar_list : list of Radar objects
        Optional. list of radar objects
    trajectory : Trajectory object
        containing trajectory samples

    Returns
    -------
    new_dataset : Trajectory object
        radar object
    ind_rad : int
        None

    """

    if procstatus != 1:
        return None, None

    if not dscfg["initialized"]:
        if trajectory is None:
            raise Exception(
                "ERROR: Undefined trajectory for dataset '%s'" % dscfg["dsname"]
            )

        if radar_list is not None:
            for radar in radar_list:
                rad = trajectory.add_radar(radar)
                trajectory.calculate_velocities(rad)
        else:
            warn("ERROR: No valid radar found")
            return None, None

        dscfg["initialized"] = True
        return trajectory, None

    return None, None


def process_traj_trt(procstatus, dscfg, radar_list=None, trajectory=None):
    """
    Processes data according to TRT trajectory

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types, can be any datatype supported by pyrad
            and available in the radar data
        time_tol : float. Dataset keyword
            tolerance between reference time of the radar volume and that of
            the TRT cell [s]. Default 100.
        alt_min, alt_max : float. Dataset keyword
            Minimum and maximum altitude of the data inside the TRT cell to
            retrieve [m MSL]. Default None
        cell_center : Bool. Dataset keyword
            If True only the range gate closest to the center of the cell is
            extracted. Default False
        latlon_tol : Float. Dataset keyword
            Tolerance in lat/lon when extracting data only from the center of
            the TRT cell. Default 0.01
    radar_list : list of Radar objects
        Optional. list of radar objects
    trajectory : Trajectory object
        containing trajectory samples

    Returns
    -------
    new_dataset : dictionary
        Dictionary containing radar_out, a radar object containing only data
        from inside the TRT cell
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    # Process
    field_names = []
    datatypes = []
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        field_names.append(get_fieldname_pyart(datatype))
        datatypes.append(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar found")
        return None, None

    # keep locally only field of interest in radar object
    radar = deepcopy(radar_list[ind_rad])
    radar.fields = dict()
    nfields_available = 0
    for field_name in field_names:
        if field_name not in radar_list[ind_rad].fields:
            warn("Datatype '%s' not available in radar data" % field_name)
            continue
        radar.add_field(field_name, radar_list[ind_rad].fields[field_name])
        nfields_available += 1

    if nfields_available == 0:
        warn("Fields not available in radar data")
        return None, None

    # get TRT cell corresponding to current radar volume
    time_tol = dscfg.get("TimeTol", 100.0)
    alt_min = dscfg.get("alt_min", None)
    alt_max = dscfg.get("alt_max", None)
    cell_center = dscfg.get("cell_center", False)
    latlon_tol = dscfg.get("latlon_tol", 0.01)  # aprox. 1 km

    inds_ray, inds_rng, lat, lon, alt = _get_gates_trt(
        radar,
        trajectory,
        dscfg["timeinfo"],
        time_tol=time_tol,
        alt_min=alt_min,
        alt_max=alt_max,
        cell_center=cell_center,
        latlon_tol=latlon_tol,
    )

    if inds_ray is None:
        return None, None

    # prepare new radar object output
    radar_roi = deepcopy(radar)

    radar_roi.range["data"] = radar.range["data"][inds_rng]
    radar_roi.ngates = inds_rng.size
    radar_roi.time["data"] = np.asarray([radar_roi.time["data"][0]])
    radar_roi.scan_type = "roi"
    radar_roi.sweep_mode["data"] = np.array(["roi"])
    radar_roi.sweep_start_ray_index["data"] = np.array([0], dtype="int32")
    radar_roi.fixed_angle["data"] = np.array([], dtype="float64")
    radar_roi.sweep_number["data"] = np.array([0], dtype="int32")
    radar_roi.nsweeps = 1

    if radar.rays_are_indexed is not None:
        radar_roi.rays_are_indexed["data"] = np.array(
            [radar.rays_are_indexed["data"][0]]
        )
    if radar.ray_angle_res is not None:
        radar_roi.ray_angle_res["data"] = np.array([radar.ray_angle_res["data"][0]])

    radar_roi.sweep_end_ray_index["data"] = np.array([1], dtype="int32")
    radar_roi.rays_per_sweep = np.array([1], dtype="int32")
    radar_roi.azimuth["data"] = np.array([], dtype="float64")
    radar_roi.elevation["data"] = np.array([], dtype="float64")
    radar_roi.nrays = 1

    radar_roi.gate_longitude["data"] = np.empty(
        (radar_roi.nrays, radar_roi.ngates), dtype=float
    )
    radar_roi.gate_latitude["data"] = np.empty(
        (radar_roi.nrays, radar_roi.ngates), dtype=float
    )
    radar_roi.gate_altitude["data"] = np.empty(
        (radar_roi.nrays, radar_roi.ngates), dtype=float
    )

    radar_roi.gate_x["data"] = np.empty(
        (radar_roi.nrays, radar_roi.ngates), dtype=float
    )
    radar_roi.gate_y["data"] = np.empty(
        (radar_roi.nrays, radar_roi.ngates), dtype=float
    )
    radar_roi.gate_z["data"] = np.empty(
        (radar_roi.nrays, radar_roi.ngates), dtype=float
    )

    radar_roi.gate_longitude["data"][0, :] = lon
    radar_roi.gate_latitude["data"][0, :] = lat
    radar_roi.gate_altitude["data"][0, :] = alt

    radar_roi.gate_x["data"][0, :] = radar.gate_x["data"][inds_ray, inds_rng]
    radar_roi.gate_y["data"][0, :] = radar.gate_y["data"][inds_ray, inds_rng]
    radar_roi.gate_z["data"][0, :] = radar.gate_z["data"][inds_ray, inds_rng]

    radar_roi.fields = dict()
    for field_name in field_names:
        if field_name not in radar.fields:
            warn("Datatype '%s' not available in radar data" % field_name)
            continue

        field_dict = deepcopy(radar.fields[field_name])
        field_dict["data"] = np.ma.empty(
            (radar_roi.nrays, radar_roi.ngates), dtype=float
        )
        field_dict["data"][0, :] = radar.fields[field_name]["data"][inds_ray, inds_rng]
        radar_roi.add_field(field_name, field_dict)

    new_dataset = {"radar_out": radar_roi}

    return new_dataset, ind_rad


def process_traj_trt_contour(procstatus, dscfg, radar_list=None, trajectory=None):
    """
    Gets the TRT cell contour corresponding to each radar volume

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types, can be any datatype supported by pyrad
            and available in the radar data
        time_tol : float. Dataset keyword
            tolerance between reference time of the radar volume and that of
            the TRT cell [s]. Default 100.
    radar_list : list of Radar objects
        Optional. list of radar objects
    trajectory : Trajectory object
        containing trajectory samples

    Returns
    -------
    new_dataset : dict
        Dictionary containing radar_out and roi_dict. Radar out is the current
        radar object. roi_dict contains the positions defining the TRT cell
        contour
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    # Process
    field_names = []
    datatypes = []
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        field_names.append(get_fieldname_pyart(datatype))
        datatypes.append(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar found")
        return None, None

    # keep locally only field of interest in radar object
    radar = deepcopy(radar_list[ind_rad])
    radar.fields = dict()
    nfields_available = 0
    for field_name in field_names:
        if field_name not in radar_list[ind_rad].fields:
            warn("Datatype '%s' not available in radar data" % field_name)
            continue
        radar.add_field(field_name, radar_list[ind_rad].fields[field_name])
        nfields_available += 1

    if nfields_available == 0:
        warn("Fields not available in radar data")
        return None, None

    # get TRT cell corresponding to current radar volume
    time_tol = dscfg.get("TimeTol", 100.0)

    roi_dict = _get_contour_trt(radar, trajectory, dscfg["timeinfo"], time_tol=time_tol)

    if roi_dict is None:
        return None, None

    new_dataset = {"radar_out": radar, "roi_dict": roi_dict}

    return new_dataset, ind_rad


def process_traj_lightning(procstatus, dscfg, radar_list=None, trajectory=None):
    """
    Return time series according to lightning trajectory

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types, can be any datatype supported by pyrad
            and available in the radar data
        data_is_log : dict. Dataset keyword
            Dictionary specifying for each field if it is in log (True) or
            linear units (False). Default False
        ang_tol : float. Dataset keyword
            Factor that multiplies the angle resolution. Used when determining
            the neighbouring rays. Default 1.2
        az_tol, el_tol : float
            azimuth and elevation tolerance (deg). Samples that have values
            beyond this tolerance from the limits in azimuth and elevation of
            the radar will be considered outside the sector. Default 3.

    radar_list : list of Radar objects
        Optional. list of radar objects
    trajectory : Trajectory object
        containing trajectory samples

    Returns
    -------
    trajectory : Trajectory object
        Object holding time series
    ind_rad : int
        radar index

    """
    if procstatus == 0:
        # first call: nothing to do
        return None, None

    if procstatus == 2:
        # last call: do the products
        if not dscfg["initialized"]:
            warn("ERROR: No trajectory dataset available!")
            return None, None
        trajdict = dscfg["traj_atplane_dict"]

        dataset = {"ts": trajdict["ts"], "final": 1}
        return dataset, trajdict["ind_rad"]

    # Process
    field_names = []
    datatypes = []
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        field_names.append(get_fieldname_pyart(datatype))
        datatypes.append(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar found")
        return None, None

    # keep locally only field of interest in radar object
    radar = deepcopy(radar_list[ind_rad])
    radar.fields = dict()
    nfields_available = 0
    for field_name in field_names:
        if field_name not in radar_list[ind_rad].fields:
            warn("Datatype '%s' not available in radar data" % field_name)
            continue
        radar.add_field(field_name, radar_list[ind_rad].fields[field_name])
        nfields_available += 1

    if nfields_available == 0:
        warn("Fields not available in radar data")
        return None, None

    ttask_start = radar.time["data"].max()
    dt_task_start = num2date(ttask_start, radar.time["units"], radar.time["calendar"])

    if not dscfg["initialized"]:
        # init
        if trajectory is None:
            raise Exception(
                "ERROR: Undefined trajectory for dataset '%s'" % dscfg["dsname"]
            )

        rad_traj = trajectory.add_radar(radar)

        description = [
            "Description:",
            "Time series of a weather radar data type at the location",
            "of the lightning flash.",
            "The time samples where the flash was out of the weather radar",
            "sector are NOT included in this file.",
            "NaN (Not a number): No data detected at the flash location.",
        ]

        ts_dict = dict()
        data_is_log = dict()
        for datatype, field_name in zip(datatypes, field_names):
            ts = TimeSeries(
                description, maxlength=trajectory.time_vector.size, datatype=datatype
            )

            unit = get_field_unit(datatype)
            name = get_field_name(datatype)
            # Append empty series: Note: sequence matters!
            ts.add_dataseries("#Flash", "", "", plot=False)
            ts.add_dataseries("Power", "", "dBm", plot=False)
            ts.add_dataseries("at_flash", name, unit, color="b")
            ts.add_dataseries("Mean", name, unit, color="r")
            ts.add_dataseries("Min", name, unit, color="k", linestyle=":")
            ts.add_dataseries("Max", name, unit, color="k", linestyle=":")
            ts.add_dataseries("#Valid", "", "", plot=False)

            ts_dict.update({field_name: ts})
            data_is_log.update({field_name: False})
            if "data_is_log" in dscfg:
                if datatype in dscfg["data_is_log"]:
                    data_is_log[field_name] = dscfg["data_is_log"] != 0
                else:
                    warn(
                        "Units type for data type "
                        + datatype
                        + " not specified. Assumed linear"
                    )

        trajdict = dict(
            {
                "radar_traj": rad_traj,
                "radar_old": None,
                "radar_old2": None,
                "last_task_start_dt": None,
                "ind_rad": ind_rad,
                "ts_dict": ts_dict,
                "data_is_log": data_is_log,
            }
        )
        traj_ind = trajectory.get_samples_in_period(end=dt_task_start)

        dscfg["traj_atplane_dict"] = trajdict
        dscfg["initialized"] = True
    else:
        # init already done
        trajdict = dscfg["traj_atplane_dict"]
        rad_traj = trajdict["radar_traj"]
        traj_ind = trajectory.get_samples_in_period(
            start=trajdict["last_task_start_dt"], end=dt_task_start
        )
        ts_dict = trajdict["ts_dict"]
        data_is_log = trajdict["data_is_log"]

    traj_time_vec = date2num(
        trajectory.time_vector, radar.time["units"], radar.time["calendar"]
    )
    if np.size(traj_ind) == 0:
        warn("No trajectory samples within current period")

        trajdict["radar_old2"] = trajdict["radar_old"]
        trajdict["radar_old"] = radar
        return None, None
    # User defined parameter
    ang_tol = dscfg.get("ang_tol", 1.2)
    az_tol = dscfg.get("az_tol", 3.0)
    el_tol = dscfg.get("el_tol", 3.0)

    az_list = []
    el_list = []
    rr_list = []
    tt_list = []
    for tind in np.nditer(traj_ind):
        az = rad_traj.azimuth_vec[tind]
        el = rad_traj.elevation_vec[tind]
        rr = rad_traj.range_vec[tind]
        tt = traj_time_vec[tind]
        dBm = trajectory.dBm[tind]
        flashnr = trajectory.flashnr_vec[tind]

        (
            radar_sel,
            traj_ray_ind,
            traj_rng_ind,
            cell_ray_inds,
            cell_rng_ind_min,
            cell_rng_ind_max,
        ) = _get_gates(
            radar,
            az,
            el,
            rr,
            tt,
            trajdict,
            ang_tol=ang_tol,
            az_tol=az_tol,
            el_tol=el_tol,
        )

        if radar_sel is None:
            continue

        # Get data samples and compute statistics
        for field_name in field_names:
            if field_name not in radar_sel.fields:
                warn("Datatype '%s' not available in radar data" % field_name)
                continue
            rdata = radar_sel.fields[field_name]["data"]
            val = rdata[traj_ray_ind, traj_rng_ind]
            cell_vals = rdata[cell_ray_inds, cell_rng_ind_min : cell_rng_ind_max + 1]

            # Compute statistics and get number of valid data
            if data_is_log[field_name]:
                val_mean = np.ma.masked
                cell_vals_valid = cell_vals.compressed()
                if cell_vals_valid.size > 0:
                    vals_lin = np.ma.power(10.0, cell_vals_valid / 10.0)
                    val_mean = np.ma.mean(vals_lin)
                    val_mean = 10.0 * np.ma.log10(val_mean)
            else:
                val_mean = np.ma.mean(cell_vals)
            val_min = np.ma.min(cell_vals)
            val_max = np.ma.max(cell_vals)

            nvals_valid = np.count_nonzero(
                np.logical_not(np.ma.getmaskarray(cell_vals))
            )

            # Add to time series dict
            ts_dict[field_name].add_timesample(
                trajectory.time_vector[tind],
                (flashnr, dBm, val, val_mean, val_min, val_max, nvals_valid),
            )

        az_list.append(az)
        el_list.append(el)
        rr_list.append(rr)
        tt_list.append(trajectory.time_vector[tind])
        # end loop over traj samples within period

    # output radar volume and flash coordinates respect to radar
    radar_sel = radar
    if trajdict["radar_old"] is not None:
        radar_sel = trajdict["radar_old"]

    dataset = {
        "azi_traj": np.asarray(az_list),
        "ele_traj": np.asarray(el_list),
        "rng_traj": np.asarray(rr_list),
        "time_traj": np.asarray(tt_list),
        "radar": radar_sel,
        "final": 0,
    }

    # update trajectory dictionary
    trajdict["last_task_start_dt"] = dt_task_start
    trajdict["radar_old2"] = trajdict["radar_old"]
    trajdict["radar_old"] = radar

    return dataset, ind_rad


def process_traj_atplane(procstatus, dscfg, radar_list=None, trajectory=None):
    """
    Return time series according to trajectory

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types, can be any datatype supported by pyrad
            and available in the radar data
        data_is_log : dict. Dataset keyword
            Dictionary specifying for each field if it is in log (True) or
            linear units (False). Default False
        ang_tol : float. Dataset keyword
            Factor that multiplies the angle resolution. Used when determining
            the neighbouring rays. Default 1.2
        az_tol, el_tol : float
            azimuth and elevation tolerance (deg). Samples that have values
            beyond this tolerance from the limits in azimuth and elevation of
            the radar will be considered outside the sector. Default 3.
        timeformat : str or None
            time format of the time series output file
    radar_list : list of Radar objects
        Optional. list of radar objects
    trajectory : Trajectory object
        containing trajectory samples

    Returns
    -------
    trajectory : Trajectory object
        Object holding time series
    ind_rad : int
        radar index

    """
    if procstatus == 0:
        # first call: nothing to do
        return None, None

    if procstatus == 2:
        # last call: do the products
        if not dscfg["initialized"]:
            warn("ERROR: No trajectory dataset available!")
            return None, None
        trajdict = dscfg["traj_atplane_dict"]

        dataset = {"ts_dict": trajdict["ts_dict"], "final": 1}
        return dataset, trajdict["ind_rad"]

    # Process
    field_names = []
    datatypes = []
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        field_names.append(get_fieldname_pyart(datatype))
        datatypes.append(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar found")
        return None, None

    # keep locally only field of interest in radar object
    radar = deepcopy(radar_list[ind_rad])
    radar.fields = dict()
    nfields_available = 0
    for field_name in field_names:
        if field_name not in radar_list[ind_rad].fields:
            warn("Datatype '%s' not available in radar data" % field_name)
            continue
        radar.add_field(field_name, radar_list[ind_rad].fields[field_name])
        nfields_available += 1

    if nfields_available == 0:
        warn("Fields not available in radar data")
        return None, None

    ttask_start = radar.time["data"].min()
    dt_task_start = num2date(ttask_start, radar.time["units"], radar.time["calendar"])

    # User defined parameter
    ang_tol = dscfg.get("ang_tol", 1.2)
    az_tol = dscfg.get("az_tol", 3.0)
    el_tol = dscfg.get("el_tol", 3.0)
    timeformat = dscfg.get("timeformat", None)

    if not dscfg["initialized"]:
        # init
        if trajectory is None:
            raise Exception(
                "ERROR: Undefined trajectory for dataset '%s'" % dscfg["dsname"]
            )

        rad_traj = trajectory.add_radar(radar)

        description = [
            "Description:",
            "Time series of a weather radar data type at the location",
            "of the plane.",
            "The time samples where the plane was out of the weather radar",
            "sector are NOT included in this file.",
            "NaN (Not a number): No rain detected at the plane location.",
        ]

        ts_dict = dict()
        data_is_log = dict()
        for datatype, field_name in zip(datatypes, field_names):
            ts = TimeSeries(
                description,
                timeformat=timeformat,
                maxlength=trajectory.time_vector.size,
                datatype=datatype,
            )

            unit = get_field_unit(datatype)
            name = get_field_name(datatype)
            # Append empty series: Note: sequence matters!
            ts.add_dataseries("at plane", name, unit, color="b")
            ts.add_dataseries("Mean", name, unit, color="r")
            ts.add_dataseries("Min", name, unit, color="k", linestyle=":")
            ts.add_dataseries("Max", name, unit, color="k", linestyle=":")
            ts.add_dataseries("#Valid", "", "", plot=False)

            ts_dict.update({field_name: ts})
            data_is_log.update({field_name: False})
            if "data_is_log" in dscfg:
                if datatype in dscfg["data_is_log"]:
                    data_is_log[field_name] = dscfg["data_is_log"] != 0
                else:
                    warn(
                        "Units type for data type "
                        + datatype
                        + " not specified. Assumed linear"
                    )

        trajdict = dict(
            {
                "radar_traj": rad_traj,
                "radar_old": None,
                "radar_old2": None,
                "last_task_start_dt": None,
                "ind_rad": ind_rad,
                "ts_dict": ts_dict,
                "data_is_log": data_is_log,
            }
        )
        traj_ind = trajectory.get_samples_in_period(end=dt_task_start)

        dscfg["traj_atplane_dict"] = trajdict
        dscfg["initialized"] = True
    else:
        # init already done
        trajdict = dscfg["traj_atplane_dict"]
        rad_traj = trajdict["radar_traj"]
        traj_ind = trajectory.get_samples_in_period(
            start=trajdict["last_task_start_dt"], end=dt_task_start
        )
        ts_dict = trajdict["ts_dict"]
        data_is_log = trajdict["data_is_log"]

    traj_time_vec = date2num(
        trajectory.time_vector, radar.time["units"], radar.time["calendar"]
    )
    if np.size(traj_ind) == 0:
        warn("No trajectory samples within current period")

        trajdict["radar_old2"] = trajdict["radar_old"]
        trajdict["radar_old"] = radar
        return None, None

    az_list = []
    el_list = []
    rr_list = []
    tt_list = []
    for tind in np.nditer(traj_ind):
        az = rad_traj.azimuth_vec[tind]
        el = rad_traj.elevation_vec[tind]
        rr = rad_traj.range_vec[tind]
        tt = traj_time_vec[tind]

        (
            radar_sel,
            traj_ray_ind,
            traj_rng_ind,
            cell_ray_inds,
            cell_rng_ind_min,
            cell_rng_ind_max,
        ) = _get_gates(
            radar,
            az,
            el,
            rr,
            tt,
            trajdict,
            ang_tol=ang_tol,
            az_tol=az_tol,
            el_tol=el_tol,
        )

        if radar_sel is None:
            continue

        # Get data samples and compute statistics
        for field_name in field_names:
            if field_name not in radar_sel.fields:
                warn("Datatype '%s' not available in radar data" % field_name)
                continue
            rdata = radar_sel.fields[field_name]["data"]
            val = rdata[traj_ray_ind, traj_rng_ind]
            cell_vals = rdata[cell_ray_inds, cell_rng_ind_min : cell_rng_ind_max + 1]

            # Compute statistics and get number of valid data
            if data_is_log[field_name]:
                val_mean = np.ma.masked
                cell_vals_valid = cell_vals.compressed()
                if cell_vals_valid.size > 0:
                    vals_lin = np.ma.power(10.0, cell_vals_valid / 10.0)
                    val_mean = np.ma.mean(vals_lin)
                    val_mean = 10.0 * np.ma.log10(val_mean)
            else:
                val_mean = np.ma.mean(cell_vals)
            val_min = np.ma.min(cell_vals)
            val_max = np.ma.max(cell_vals)

            nvals_valid = np.count_nonzero(
                np.logical_not(np.ma.getmaskarray(cell_vals))
            )

            # Add to time series dict
            ts_dict[field_name].add_timesample(
                trajectory.time_vector[tind],
                (val, val_mean, val_min, val_max, nvals_valid),
            )

        az_list.append(az)
        el_list.append(el)
        rr_list.append(rr)
        tt_list.append(trajectory.time_vector[tind])
        # end loop over traj samples within period

    # output radar volume and antenna coordinates respect to radar
    radar_sel = radar
    if trajdict["radar_old"] is not None:
        radar_sel = trajdict["radar_old"]

    dataset = {
        "azi_traj": np.asarray(az_list),
        "ele_traj": np.asarray(el_list),
        "rng_traj": np.asarray(rr_list),
        "time_traj": np.asarray(tt_list),
        "radar": radar_sel,
        "final": 0,
    }

    # update trajectory dictionary
    trajdict["last_task_start_dt"] = dt_task_start
    trajdict["radar_old2"] = trajdict["radar_old"]
    trajdict["radar_old"] = radar

    return dataset, ind_rad


def process_traj_antenna_pattern(procstatus, dscfg, radar_list=None, trajectory=None):
    """
    Process a new array of data volumes considering a plane
    trajectory. As result a timeseries with the values
    transposed for a given antenna pattern is created.
    The result is created when the LAST flag is set.

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        datatype : list of string. Dataset keyword
            The input data types, can be any datatype supported by pyrad
            and available in the radar data
        antennaType : str. Dataset keyword
            Type of antenna of the radar we want to get the view from. Can
            be AZIMUTH, ELEVATION, LOWBEAM, HIGHBEAM
        par_azimuth_antenna : dict. Global ekyword
            Dictionary containing the parameters of the PAR azimuth antenna,
            i.e. name of the file with the antenna elevation pattern and fixed
            antenna angle
        par_elevation_antenna : dict. Global keyword
            Dictionary containing the parameters of the PAR elevation antenna,
            i.e. name of the file with the antenna azimuth pattern and fixed
            antenna angle
        asr_lowbeam_antenna : dict. Global keyword
            Dictionary containing the parameters of the ASR low beam antenna,
            i.e. name of the file with the antenna elevation pattern and fixed
            antenna angle
        asr_highbeam_antenna : dict. Global keyword
            Dictionary containing the parameters of the ASR high beam antenna,
            i.e. name of the file with the antenna elevation pattern and fixed
            antenna angle
        target_radar_pos : dict. Global keyword
            Dictionary containing the latitude, longitude and altitude of
            the radar we want to get the view from. If not specifying it will
            assume the radar is collocated
        range_all : Bool. Dataset keyword
            If the real radar and the synthetic radar are co-located and this
            parameter is true the statistics are going to be computed using
            all the data from range 0 to the position of the plane. Default
            False
        rhi_resolution : Bool. Dataset keyword
            Resolution of the synthetic RHI used to compute the data as viewed
            from the synthetic radar [deg]. Default 0.5
        max_altitude : float. Dataset keyword
            Max altitude of the data to use when computing the view from the
            synthetic radar [m MSL]. Default 12000.
        latlon_tol : float. Dataset keyword
            The tolerance in latitude and longitude to determine which
            synthetic radar gates are co-located with real radar gates [deg].
            Default 0.04
        alt_tol : float. Datset keyword
            The tolerance in altitude to determine which synthetic
            radar gates are co-located with real radar gates [m]. Default 1000.
        distance_upper_bound : float. Dataset keyword
            The maximum distance where to look for a neighbour when
            determining which synthetic radar gates are co-located with real
            radar gates [m]. Default 1000.
        use_cKDTree : Bool. Dataset keyword
            Which function to use to find co-located real radar gates with the
            synthetic radar. If True a function using cKDTree from
            scipy.spatial is used. This function uses parameter
            distance_upper_bound. If False a native implementation is used
            that takes as parameters latlon_tol and alt_tol. Default True.
        pattern_thres : float. Dataset keyword
            The minimum of the sum of the weights given to each value in order
            to consider the weighted quantile valid. It is related to the
            number of valid data points
        data_is_log : dict. Dataset keyword
            Dictionary specifying for each field if it is in log (True) or
            linear units (False). Default False
        use_nans : dict. Dataset keyword
            Dictionary specyfing whether the nans have to be used in the
            computation of the statistics for each field. Default False
        nan_value : dict. Dataset keyword
            Dictionary with the value to use to substitute the NaN values when
            computing the statistics of each field. Default 0
        az_tol, el_tol : float
            azimuth and elevation tolerance (deg). Samples that have values
            beyond this tolerance from the limits in azimuth and elevation of
            the radar will be considered outside the sector. Default 3.
    radar_list : list of Radar objects
        Optional. list of radar objects
    trajectory : Trajectory object
        containing trajectory samples

    Returns
    -------
    trajectory : Trajectory object
        Object holding time series
    ind_rad : int
        radar index
    """

    if procstatus == 0:
        # first call: nothing to do
        return None, None

    if procstatus == 2:
        # last call: do the products
        if not dscfg["initialized"]:
            warn("ERROR: No trajectory dataset available!")
            return None, None
        tadict = dscfg["traj_antenna_dict"]
        dataset = {"ts_dict": tadict["ts_dict"], "final": 1}
        return dataset, tadict["ind_rad"]

    # Process
    field_names = []
    datatypes = []
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        field_names.append(get_fieldname_pyart(datatype))
        datatypes.append(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar found")
        return None, None

    # keep locally only field of interest in radar object
    radar = deepcopy(radar_list[ind_rad])
    radar.fields = dict()
    nfields_available = 0
    for field_name in field_names:
        if field_name not in radar_list[ind_rad].fields:
            warn("Datatype '%s' not available in radar data" % field_name)
            continue
        radar.add_field(field_name, radar_list[ind_rad].fields[field_name])
        nfields_available += 1

    if nfields_available == 0:
        warn("Fields not available in radar data")
        return None, None

    ttask_start = radar.time["data"].min()
    dt_task_start = num2date(ttask_start, radar.time["units"], radar.time["calendar"])
    if not dscfg["initialized"]:
        # === init ============================================================
        if trajectory is None:
            raise Exception(
                "ERROR: Undefined trajectory for dataset '%s'" % dscfg["dsname"]
            )

        # Check config
        if "antennaType" not in dscfg:
            raise Exception(
                "ERROR: Undefined 'antennaType' for dataset '%s'" % dscfg["dsname"]
            )
        if "configpath" not in dscfg:
            raise Exception(
                "ERROR: Undefined 'configpath' for dataset '%s'" % dscfg["dsname"]
            )

        if "target_radar_pos" not in dscfg:
            radar_antenna_atsameplace = True
            target_radar = None

            warn(
                "No target radar position specified. "
                + "The radars are assumed colocated"
            )
            rad_traj = trajectory.add_radar(radar)
        else:
            radar_antenna_atsameplace = False

            # create dummy radar object with target radar specs
            latitude = get_metadata("latitude")
            longitude = get_metadata("longitude")
            altitude = get_metadata("altitude")

            latitude["data"] = np.array(
                [dscfg["target_radar_pos"]["latitude"]], dtype="float64"
            )
            longitude["data"] = np.array(
                [dscfg["target_radar_pos"]["longitude"]], dtype="float64"
            )
            altitude["data"] = np.array(
                [dscfg["target_radar_pos"]["altitude"]], dtype="float64"
            )

            target_radar = TargetRadar(latitude, longitude, altitude)
            rad_traj = trajectory.add_radar(target_radar)

        if dscfg["antennaType"] == "AZIMUTH":
            is_azimuth_antenna = True
            info = "parAzAnt"
            description = [
                "Antenna: PAR Azimuth antenna",
                "Description:",
                "Time series of a weather radar data type at the location",
                "of the plane weighted by the antenna pattern of the PAR",
                "antenna.",
                "The time samples where the plane was out of the weather",
                "radar sector are NOT included in this file.",
                "NaN (Not a number): No rain detected at the plane location.",
            ]
            if "par_azimuth_antenna" not in dscfg:
                raise Exception(
                    "ERROR: Undefined 'par_azimuth_antenna' for"
                    " dataset '%s'" % dscfg["dsname"]
                )

            patternfile = (
                dscfg["configpath"]
                + "antenna/"
                + dscfg["par_azimuth_antenna"]["elPatternFile"]
            )
            fixed_angle = dscfg["par_azimuth_antenna"]["fixed_angle"]

        elif dscfg["antennaType"] == "ELEVATION":
            is_azimuth_antenna = False
            info = "parElAnt"
            description = [
                "Antenna: PAR Elevation antenna",
                "Description:",
                "Time series of a weather radar data type at the location",
                "of the plane weighted by the antenna pattern of the PAR",
                "antenna.",
                "The time samples where the plane was out of the weather ",
                "radar sector are NOT included in this file.",
                "NaN (Not a number): No rain detected at the plane location.",
            ]

            if "par_elevation_antenna" not in dscfg:
                raise Exception(
                    "ERROR: Undefined 'par_elevation_antenna' for"
                    " dataset '%s'" % dscfg["dsname"]
                )

            patternfile = (
                dscfg["configpath"]
                + "antenna/"
                + dscfg["par_elevation_antenna"]["azPatternFile"]
            )
            fixed_angle = dscfg["par_elevation_antenna"]["fixed_angle"]

        elif dscfg["antennaType"] == "LOWBEAM":
            is_azimuth_antenna = True
            info = "asrLowBeamAnt"
            description = [
                "Antenna: ASR low beam antenna",
                "Description:",
                "Time series of a weather radar data type at the location",
                "of the plane weighted by the antenna pattern of the ASR",
                "antenna.",
                "The time samples where the plane was out of the weather",
                "radar sector are NOT included in this file.",
                "NaN (Not a number): No rain detected at the plane location.",
            ]
            if "asr_lowbeam_antenna" not in dscfg:
                raise Exception(
                    "ERROR: Undefined 'asr_lowbeam_antenna' for"
                    " dataset '%s'" % dscfg["dsname"]
                )

            patternfile = (
                dscfg["configpath"]
                + "antenna/"
                + dscfg["asr_lowbeam_antenna"]["elPatternFile"]
            )
            fixed_angle = dscfg["asr_lowbeam_antenna"]["fixed_angle"]

        elif dscfg["antennaType"] == "HIGHBEAM":
            is_azimuth_antenna = True
            info = "asrHighBeamAnt"
            description = [
                "Antenna: ASR high beam antenna",
                "Description:",
                "Time series of a weather radar data type at the location",
                "of the plane weighted by the antenna pattern of the ASR",
                "antenna.",
                "The time samples where the plane was out of the weather",
                "radar sector are NOT included in this file.",
                "NaN (Not a number): No rain detected at the plane location.",
            ]
            if "asr_highbeam_antenna" not in dscfg:
                raise Exception(
                    "ERROR: Undefined 'asr_highbeam_antenna' for"
                    " dataset '%s'" % dscfg["dsname"]
                )

            patternfile = (
                dscfg["configpath"]
                + "antenna/"
                + dscfg["asr_highbeam_antenna"]["elPatternFile"]
            )
            patternfile_low = (
                dscfg["configpath"]
                + "antenna/"
                + dscfg["asr_lowbeam_antenna"]["elPatternFile"]
            )
            fixed_angle = dscfg["asr_highbeam_antenna"]["fixed_angle"]
        else:
            raise Exception(
                "ERROR: Unexpected antenna type '%s' for dataset"
                " '%s'" % (dscfg["antennaType"], dscfg["dsname"])
            )

        # Read dataset config parameters:
        weight_threshold = dscfg.get("pattern_thres", 0.0)

        do_all_ranges = False
        if "range_all" in dscfg:
            if dscfg["range_all"] != 0:
                do_all_ranges = True

        # Config parameters for processing when the weather radar and the
        # antenna are not at the same place:
        rhi_resolution = dscfg.get("rhi_resolution", 0.5)  # [deg]
        max_altitude = dscfg.get("max_altitude", 12000.0)  # [m]
        latlon_tol = dscfg.get("latlon_tol", 0.04)  # [deg]
        alt_tol = dscfg.get("alt_tol", 1000.0)  # [m]
        distance_upper_bound = dscfg.get("distance_upper_bound", 1000.0)
        use_cKDTree = dscfg.get("use_cKDTree", True)

        # Get antenna pattern and make weight vector
        try:
            if info == "asrHighBeamAnt":
                antpattern = read_antenna_pattern(
                    patternfile, linear=True, twoway=False
                )
                antpattern_low = read_antenna_pattern(
                    patternfile_low, linear=True, twoway=False
                )
                antpattern["attenuation"] *= antpattern_low["attenuation"]
            else:
                antpattern = read_antenna_pattern(patternfile, linear=True, twoway=True)
        except Exception as ee:
            warn(str(ee))
            raise

        pattern_angles = antpattern["angle"] + fixed_angle
        if not is_azimuth_antenna:
            pattern_angles[pattern_angles < 0] += 360.0
        pattern_angles[pattern_angles >= 360.0] -= 360.0

        if radar_antenna_atsameplace:
            if is_azimuth_antenna:
                scan_angles = np.sort(
                    np.unique(radar.elevation["data"].round(decimals=1))
                )
            else:
                scan_angles = np.sort(
                    np.unique(radar.azimuth["data"].round(decimals=1))
                )
        else:
            scan_angles = np.arange(0, 90, rhi_resolution, dtype=float)

        weightvec = np.empty(scan_angles.size, dtype=float)
        for kk in range(scan_angles.size):
            ind = np.argmin(np.abs(pattern_angles - scan_angles[kk]))
            weightvec[kk] = antpattern["attenuation"][ind]

        ts_dict = dict()
        data_is_log = dict()
        use_nans = dict()
        nan_value = dict()
        for datatype, field_name in zip(datatypes, field_names):
            ts = TimeSeries(
                description, maxlength=trajectory.time_vector.size, datatype=datatype
            )

            unit = get_field_unit(datatype)
            name = get_field_name(datatype)
            # Quantiles of interest
            quantiles = [
                {"val": 0.1, "plot": False, "color": None, "ltype": None},
                {"val": 0.2, "plot": True, "color": "k", "ltype": ":"},
                {"val": 0.3, "plot": False, "color": None, "ltype": None},
                {"val": 0.4, "plot": False, "color": None, "ltype": None},
                {"val": 0.5, "plot": True, "color": "r", "ltype": None},
                {"val": 0.6, "plot": False, "color": None, "ltype": None},
                {"val": 0.7, "plot": False, "color": None, "ltype": None},
                {"val": 0.8, "plot": True, "color": "k", "ltype": ":"},
                {"val": 0.9, "plot": False, "color": None, "ltype": None},
                {"val": 0.95, "plot": False, "color": None, "ltype": None},
            ]

            ts.add_dataseries("Weighted average", name, unit, color="b")
            for qq in quantiles:
                label = "Quantile_%4.2f" % qq["val"]
                ts.add_dataseries(
                    label,
                    name,
                    unit,
                    plot=qq["plot"],
                    color=qq["color"],
                    linestyle=qq["ltype"],
                )
            ts.add_dataseries("#Valid", "", "", plot=False)
            ts_dict.update({field_name: ts})

            data_is_log.update({field_name: False})
            if "data_is_log" in dscfg:
                if datatype in dscfg["data_is_log"]:
                    data_is_log[field_name] = dscfg["data_is_log"][datatype] != 0
                else:
                    warn(
                        "Units type for data type "
                        + datatype
                        + " not specified. Assumed linear"
                    )

            use_nans.update({field_name: False})
            if "use_nans" in dscfg:
                if datatype in dscfg["use_nans"]:
                    use_nans[field_name] = dscfg["use_nans"][datatype] != 0
                else:
                    warn(
                        "Use of nans not specified for data type "
                        + datatype
                        + " not specified. Assumed not used"
                    )

            nan_value.update({field_name: 0.0})
            if "nan_value" in dscfg:
                if datatype in dscfg["nan_value"]:
                    nan_value[field_name] = dscfg["nan_value"][datatype]
                else:
                    warn(
                        "NaN value not specified for data type "
                        + datatype
                        + " not specified. Assumed 0"
                    )

        quants = np.array([ee["val"] for ee in quantiles])

        # Persistent data structure
        tadict = dict(
            {
                "radar_traj": rad_traj,
                "radar_old": None,
                "radar_old2": None,
                "target_radar": target_radar,
                "last_task_start_dt": None,
                "ind_rad": ind_rad,
                "is_azimuth_antenna": is_azimuth_antenna,
                "info": info,
                "scan_angles": scan_angles,
                "radar_antenna_atsameplace": radar_antenna_atsameplace,
                "weightvec": weightvec,
                "quantiles": quants,
                "use_nans": use_nans,
                "nan_value": nan_value,
                "weight_threshold": weight_threshold,
                "do_all_ranges": do_all_ranges,
                "max_altitude": max_altitude,
                "latlon_tol": latlon_tol,
                "alt_tol": alt_tol,
                "distance_upper_bound": distance_upper_bound,
                "use_cKDTree": use_cKDTree,
                "data_is_log": data_is_log,
                "ts_dict": ts_dict,
            }
        )
        traj_ind = trajectory.get_samples_in_period(end=dt_task_start)

        dscfg["traj_antenna_dict"] = tadict
        dscfg["initialized"] = True
        # end init
    else:
        # init already done
        tadict = dscfg["traj_antenna_dict"]

        traj_ind = trajectory.get_samples_in_period(
            start=tadict["last_task_start_dt"], end=dt_task_start
        )

    az_tol = dscfg.get("az_tol", 3.0)
    el_tol = dscfg.get("el_tol", 3.0)

    if not _get_ts_values_antenna_pattern(
        radar, trajectory, tadict, traj_ind, field_names, az_tol=az_tol, el_tol=el_tol
    ):
        return None, None

    tadict["last_task_start_dt"] = dt_task_start
    tadict["radar_old2"] = tadict["radar_old"]
    tadict["radar_old"] = radar

    # Collect garbage
    gc.collect()

    return None, None


def _get_ts_values_antenna_pattern(
    radar, trajectory, tadict, traj_ind, field_names, az_tol=3.0, el_tol=3.0
):
    """
    Get the time series values of a trajectory using a synthetic antenna
    pattern

    Parameters
    ----------
    radar : radar object
        The radar volume with the data
    trajectory : trajectory object
        The plane trajectory
    tadict : dict
        A dictionary containing parameters useful for trajectory computation
    traj_ind : array
        The indices of trajectory data within the current radar volume time
    field_names : list of str
        list of names of the radar field
    az_tol, el_tol : float
        azimuth and elevation tolerance (deg). Samples that have values beyond
        this tolerance from the limits in azimuth and elevation of the radar
        will be considered outside the sector

    Returns
    -------
    result : Bool
        A flag signaling whether radar data matching the trajectory was found

    """
    rad_traj = tadict["radar_traj"]
    is_azimuth_antenna = tadict["is_azimuth_antenna"]
    scan_angles = tadict["scan_angles"]
    radar_antenna_atsameplace = tadict["radar_antenna_atsameplace"]
    nan_value = tadict["nan_value"]
    use_nans = tadict["use_nans"]
    weight_threshold = tadict["weight_threshold"]
    do_all_ranges = tadict["do_all_ranges"]
    ts_dict = tadict["ts_dict"]
    target_radar = tadict["target_radar"]
    max_altitude = tadict["max_altitude"]
    latlon_tol = tadict["latlon_tol"]
    alt_tol = tadict["alt_tol"]
    distance_upper_bound = tadict["distance_upper_bound"]
    use_cKDTree = tadict["use_cKDTree"]
    data_is_log = tadict["data_is_log"]

    traj_time_vec = date2num(
        trajectory.time_vector, radar.time["units"], radar.time["calendar"]
    )

    if np.size(traj_ind) == 0:
        warn("No trajectory samples within current period")
        return False

    for tind in np.nditer(traj_ind):
        az = rad_traj.azimuth_vec[tind]
        el = rad_traj.elevation_vec[tind]
        rr = rad_traj.range_vec[tind]
        tt = traj_time_vec[tind]

        # Select radar object, find closest azimuth and elevation ray
        (radar_sel, ray_sel, rr_ind, el_vec_rnd, az_vec_rnd) = _get_closest_bin(
            az, el, rr, tt, radar, tadict
        )

        # Check if traj sample is within scan sector
        if _sample_out_of_sector(
            az,
            el,
            rr,
            radar_sel,
            ray_sel,
            rr_ind,
            el_vec_rnd,
            az_vec_rnd,
            az_tol=az_tol,
            el_tol=el_tol,
        ):
            continue

        if radar_antenna_atsameplace:
            # ==============================================================
            # Radar and scanning antenna are at the SAME place
            # ==============================================================

            # ==============================================================
            # Get sample at bin

            if is_azimuth_antenna:
                angles = radar_sel.azimuth["data"]
                angles_scan = radar_sel.elevation["data"]
                ray_angle = radar_sel.azimuth["data"][ray_sel]
            else:
                angles = radar_sel.elevation["data"]
                angles_scan = radar_sel.azimuth["data"]
                ray_angle = radar_sel.elevation["data"][ray_sel]

            d_angle = np.abs(angles - ray_angle)
            ray_inds = np.where(d_angle < 0.09)[0]
            angles_sortind = np.argsort(angles_scan[ray_inds])

            ray_inds = ray_inds[angles_sortind]
            angles_sorted = angles_scan[ray_inds]

            # Set default values
            avg = None
            qvals = np.array([None] * tadict["quantiles"].size)
            nvals_valid = None

            if (scan_angles.size != angles_sorted.size) or (
                np.max(np.abs(scan_angles - angles_sorted)) > 0.1
            ):
                warn("Scan angle mismatch!")
                for field_name in field_names:
                    ts_dict[field_name].add_timesample(
                        trajectory.time_vector[tind],
                        np.concatenate([[avg], qvals, [nvals_valid]]),
                    )
                continue

            if do_all_ranges:
                rr_ind_min = 0
            else:
                rr_ind_min = rr_ind

            w_vec = tadict["weightvec"]
            for field_name in field_names:
                if field_name not in radar_sel.fields:
                    warn("Datatype '%s' not available in radar data" % field_name)
                    continue
                rdata = radar_sel.fields[field_name]["data"]
                values = rdata[ray_inds, rr_ind_min : rr_ind + 1]
                if use_nans[field_name]:
                    values_ma = np.ma.getmaskarray(values)
                    values[values_ma] = nan_value[field_name]
                try:
                    (avg, qvals, nvals_valid) = quantiles_weighted(
                        values,
                        weight_vector=w_vec,
                        quantiles=tadict["quantiles"],
                        weight_threshold=weight_threshold,
                        data_is_log=data_is_log[field_name],
                    )
                except Exception as ee:
                    warn(str(ee))
                    continue

                ts_dict[field_name].add_timesample(
                    trajectory.time_vector[tind],
                    np.concatenate([[avg], qvals, [nvals_valid]]),
                )
        else:
            # ================================================================
            # Radar and scanning antenna are NOT at the same place
            # ================================================================
            ray_inds, rng_inds, w_inds = _get_gates_antenna_pattern(
                radar_sel,
                target_radar,
                az,
                rr,
                tt,
                scan_angles,
                alt_tol=alt_tol,
                latlon_tol=latlon_tol,
                max_altitude=max_altitude,
                distance_upper_bound=distance_upper_bound,
                use_cKDTree=use_cKDTree,
            )

            w_vec = tadict["weightvec"][w_inds]

            for field_name in field_names:
                if field_name not in radar_sel.fields:
                    warn("Datatype '%s' not available in radar data" % field_name)
                    continue
                rdata = radar_sel.fields[field_name]["data"]
                values = rdata[ray_inds, rng_inds]
                if use_nans[field_name]:
                    values_ma = np.ma.getmaskarray(values)
                    values[values_ma] = nan_value[field_name]

                try:
                    (avg, qvals, nvals_valid) = quantiles_weighted(
                        values,
                        weight_vector=w_vec,
                        quantiles=tadict["quantiles"],
                        weight_threshold=weight_threshold,
                        data_is_log=data_is_log[field_name],
                    )
                except Exception as ee:
                    warn(str(ee))
                    continue

                ts_dict[field_name].add_timesample(
                    trajectory.time_vector[tind],
                    np.concatenate([[avg], qvals, [nvals_valid]]),
                )
    return True


def _get_contour_trt(radar, trajectory, voltime, time_tol=100.0):
    """
    Get the TRT cell contour corresponding to the current radar

    Parameters
    ----------
    radar : radar object
        The radar containing
    trajectory : trajectory object
        Object containing the TRT cell position and dimensions
    voltime : datetime object
        The radar volume reference time
    time_tol : float
        Time tolerance where to look for data [s]

    Returns
    -------
    lon_roi, lat_roi : array of floats
        The lat/lon defining the TRT cell contour

    """
    dt = np.empty(trajectory.time_vector.size, dtype=float)
    for i, time_traj in enumerate(trajectory.time_vector):
        dt[i] = np.abs((voltime - time_traj).total_seconds())
    if dt.min() > time_tol:
        warn("No TRT data for radar volume time {}".format(voltime))
        return None
    ind = np.argmin(dt)

    lon_roi = trajectory.cell_contour[ind]["lon"]
    lat_roi = trajectory.cell_contour[ind]["lat"]
    lat_center = trajectory.wgs84_lat_deg[ind]
    lon_center = trajectory.wgs84_lon_deg[ind]

    roi_dict = {
        "lon": lon_roi,
        "lat": lat_roi,
        "lon_center": lon_center,
        "lat_center": lat_center,
    }

    # extract the data within the ROI boundaries
    inds_ray, inds_rng = np.indices(np.shape(radar.gate_longitude["data"]))

    mask = np.logical_and(
        np.logical_and(
            radar.gate_latitude["data"] >= roi_dict["lat"].min(),
            radar.gate_latitude["data"] <= roi_dict["lat"].max(),
        ),
        np.logical_and(
            radar.gate_longitude["data"] >= roi_dict["lon"].min(),
            radar.gate_longitude["data"] <= roi_dict["lon"].max(),
        ),
    )

    if np.all(mask == 0):
        warn("No values within ROI")
        return None

    inds_ray = inds_ray[mask]
    inds_rng = inds_rng[mask]

    # extract the data inside the ROI
    lat = radar.gate_latitude["data"][mask]
    lon = radar.gate_longitude["data"][mask]
    inds, is_roi = belongs_roi_indices(lat, lon, roi_dict)

    if is_roi == "None":
        warn("No values within ROI")
        return None

    return roi_dict


def _get_gates(radar, az, el, rr, tt, trajdict, ang_tol=1.2, az_tol=3.0, el_tol=3.0):
    """
    Find the gates of the radar object that have to be used to compute the
    data of a trajectory

    Parameters
    ----------
    radar : radar object
        The radar containing
    az, el, rr : floats
        The trajectory position respect to the radar
    tt : float
        the trajectory time respect to the beginning of the radar scan
    trajdict : dict
        Dictionary containing the trajectory parameters
    ang_tol : float
        Factor that multiplies the angle resolution. Used when determining
        the neighbouring rays
    az_tol, el_tol : float
        azimuth and elevation tolerance (deg). Samples that have values beyond
        this tolerance from the limits in azimuth and elevation of the radar
        will be considered outside the sector

    Returns
    -------
    radar_sel : radar object
        The radar volume selected as closest to trajectory point
    ray_sel, rr_ind : ints
        ray and range indices of the radar gate closest to the trajectory
        position
    cell_ind : array of ints
        indices of the surrounding rays
    rr_min : int
        index of the minimum range of the surrounding gates
    rr_max : int
        index of the maximum range of the surrounding gates

    """
    # Find closest azimuth and elevation ray
    (radar_sel, ray_sel, rr_ind, el_vec_rnd, az_vec_rnd) = _get_closest_bin(
        az, el, rr, tt, radar, trajdict
    )

    # Check if traj sample is within scan sector
    if _sample_out_of_sector(
        az,
        el,
        rr,
        radar_sel,
        ray_sel,
        rr_ind,
        el_vec_rnd,
        az_vec_rnd,
        az_tol=az_tol,
        el_tol=el_tol,
    ):
        return None, None, None, None, None, None

    # Get indices of gates surrounding the cell (3x3 box)
    el_res = np.ma.median(np.ma.diff(el_vec_rnd))
    az_res = np.ma.median(np.ma.diff(az_vec_rnd))

    d_el = np.abs(radar_sel.elevation["data"] - radar_sel.elevation["data"][ray_sel])
    d_az = np.abs(radar_sel.azimuth["data"] - radar_sel.azimuth["data"][ray_sel])
    cell_ind = np.where((d_el < el_res * ang_tol) & (d_az < az_res * ang_tol))

    rr_min = rr_ind - 1
    if rr_min < 0:
        rr_min = 0

    return radar_sel, ray_sel, rr_ind, cell_ind, rr_min, rr_ind + 1


def _get_gates_trt(
    radar,
    trajectory,
    voltime,
    time_tol=100.0,
    alt_min=None,
    alt_max=None,
    cell_center=False,
    latlon_tol=0.0005,
):
    """
    Find the gates of the radar object that belong to a TRT cell

    Parameters
    ----------
    radar : radar object
        The radar containing
    trajectory : trajectory object
        Object containing the TRT cell position and dimensions
    voltime : datetime object
        The radar volume reference time
    time_tol : float
        Time tolerance where to look for data [s]
    alt_min, alt_max : float
        Minimum and maximum altitude where to look for data [m]

    Returns
    -------
    inds_ray, inds_rng : array of ints
        The indices of the radar data inside the TRT cell

    """
    dt = np.empty(trajectory.time_vector.size, dtype=float)
    for i, time_traj in enumerate(trajectory.time_vector):
        dt[i] = np.abs((voltime - time_traj).total_seconds())
    if dt.min() > time_tol:
        warn("No TRT data for radar volume time")
        return None, None, None, None, None
    ind = np.argmin(dt)

    if not cell_center:
        lon_roi = trajectory.cell_contour[ind]["lon"]
        lat_roi = trajectory.cell_contour[ind]["lat"]

        roi_dict = {
            "lon": lon_roi,
            "lat": lat_roi,
            "alt_min": alt_min,
            "alt_max": alt_max,
        }

        # extract the data within the ROI boundaries
        inds_ray, inds_rng = np.indices(np.shape(radar.gate_longitude["data"]))

        mask = np.logical_and(
            np.logical_and(
                radar.gate_latitude["data"] >= roi_dict["lat"].min(),
                radar.gate_latitude["data"] <= roi_dict["lat"].max(),
            ),
            np.logical_and(
                radar.gate_longitude["data"] >= roi_dict["lon"].min(),
                radar.gate_longitude["data"] <= roi_dict["lon"].max(),
            ),
        )

        if alt_min is not None:
            mask[radar.gate_altitude["data"] < alt_min] = 0
        if alt_max is not None:
            mask[radar.gate_altitude["data"] > alt_max] = 0

        if np.all(mask == 0):
            warn("No values within ROI")
            return None, None, None, None, None

        inds_ray = inds_ray[mask]
        inds_rng = inds_rng[mask]

        # extract the data inside the ROI
        lat = radar.gate_latitude["data"][mask]
        lon = radar.gate_longitude["data"][mask]
        inds, is_roi = belongs_roi_indices(lat, lon, roi_dict)

        if is_roi == "None":
            warn("No values within ROI")
            return None, None, None, None, None

        inds_ray = inds_ray[inds]
        inds_rng = inds_rng[inds]

        lat = lat[inds]
        lon = lon[inds]
        alt = radar.gate_altitude["data"][inds_ray, inds_rng]
    else:
        radar_aux = deepcopy(radar)
        # transform radar into ppi over the required elevation
        if radar_aux.scan_type == "rhi":
            target_elevations, el_tol = get_target_elevations(radar_aux)
            radar_ppi = cross_section_rhi(radar_aux, target_elevations, el_tol=el_tol)
        elif radar_aux.scan_type == "ppi":
            radar_ppi = radar_aux
        else:
            warn("Error: unsupported scan type.")
            return None, None, None, None, None

        inds_ray = np.array([], dtype=int)
        inds_rng = np.array([], dtype=int)
        lat = np.array([])
        lon = np.array([])
        alt = np.array([])
        for sweep in range(radar_ppi.nsweeps):
            radar_aux = deepcopy(radar_ppi.extract_sweeps([sweep]))

            # find nearest gate to lat lon point
            ind_ray, ind_rng, _, _ = find_nearest_gate(
                radar_aux,
                trajectory.wgs84_lat_deg[ind],
                trajectory.wgs84_lon_deg[ind],
                latlon_tol=latlon_tol,
            )

            if ind_ray is None:
                continue

            if alt_min is not None:
                if radar_aux.gate_altitude["data"][ind_ray][ind_rng] < alt_min:
                    continue
            if alt_max is not None:
                if radar_aux.gate_altitude["data"][ind_ray][ind_rng] > alt_max:
                    continue

            inds_ray = np.append(inds_ray, ind_ray)
            inds_rng = np.append(inds_rng, ind_rng)
            lat = np.append(lat, radar_aux.gate_latitude["data"][ind_ray][ind_rng])
            lon = np.append(lon, radar_aux.gate_longitude["data"][ind_ray][ind_rng])
            alt = np.append(alt, radar_aux.gate_altitude["data"][ind_ray][ind_rng])

        if inds_ray.size == 0:
            warn("No values in center of cell")
            return None, None, None, None, None

    return inds_ray, inds_rng, lat, lon, alt


def _get_gates_antenna_pattern(
    radar_sel,
    target_radar,
    az,
    rr,
    tt,
    scan_angles,
    alt_tol=1000.0,
    latlon_tol=0.04,
    max_altitude=12000.0,
    distance_upper_bound=1000.0,
    use_cKDTree=True,
):
    """
    Find the gates of the radar object that have to be used to compute the
    data of a trajectory as seen by another radar system

    Parameters
    ----------
    radar_sel : radar object
        The radar containing real data
    target_radar : radar object
        The virtual radar
    az, rr : floats
        The trajectory position respect to the radar
    tt : float
        the trajectory time respect to the beginning of the radar scan
    scan_angles : array
        The scan angles of the virtual radar object
    alt_tol : float
        The tolerance in altitude [m]
    latlon_tol : float
        The tolerance in latitude and longitude [deg]
    max_altitude : float
        The maximum altitude where to look for radar data [m MSL]
    distance_upper_bound : float
        The maximum distance where to look for neighbors [m]
    use_cKDTree : Bool
        If True the nearest neighbour to the synthetic radar is found using
        the cKDTree function of scipy.spatial. Otherwise it is found by Pyrad
        implementation

    Returns
    -------
    ray_ind, rng_ind : array of ints
        the indices of the radar data to use
    w_ind : array of ints
        The indices of the one-dimensional antenna pattern to use

    """
    # Create dummy Radar object with fix az and r, and all elevations
    n_rays = scan_angles.size

    r_time = {"data": np.array([tt * n_rays])}
    r_range = {"data": np.array([rr])}
    r_azimuth = get_metadata("azimuth")
    r_azimuth["data"] = np.ones(n_rays) * az
    r_elevation = {"data": scan_angles}
    r_sweep_number = {"data": np.array([0])}
    r_fields = {"colocated_gates": get_metadata("colocated_gates")}
    r_fields["colocated_gates"]["data"] = 2 * np.ma.ones((n_rays, 1), dtype=int)

    r_radar = Radar(
        r_time,
        r_range,
        r_fields,
        None,
        "rhi",
        target_radar.latitude,
        target_radar.longitude,
        target_radar.altitude,
        r_sweep_number,
        None,
        None,
        None,
        None,
        r_azimuth,
        r_elevation,
    )

    # flag regions with colocated usable data in r_radar
    r_ind_invalid = r_radar.gate_altitude["data"] > max_altitude
    r_radar.fields["colocated_gates"]["data"][r_ind_invalid] = 1

    # flag regions with colocated usable data in radar_sel
    gate_coloc_radar_sel = pyart.util.intersection(
        radar_sel,
        r_radar,
        h_tol=alt_tol,
        latlon_tol=latlon_tol,
        vol_d_tol=None,
        vismin=None,
        hmin=None,
        hmax=max_altitude,
        rmin=None,
        rmax=None,
        elmin=None,
        elmax=None,
        azmin=None,
        azmax=None,
        visib_field=None,
        intersec_field="colocated_gates",
    )
    radar_sel.add_field("colocated_gates", gate_coloc_radar_sel, replace_existing=True)

    if use_cKDTree:
        colgates, r_radar_colg = pyart.util.colocated_gates2(
            r_radar, radar_sel, distance_upper_bound=distance_upper_bound
        )
    else:
        colgates, r_radar_colg = pyart.util.colocated_gates(
            r_radar, radar_sel, h_tol=alt_tol, latlon_tol=latlon_tol
        )

    w_ind = np.where(r_radar_colg["data"] > 1)[0]

    return colgates["rad2_ray_ind"], colgates["rad2_rng_ind"], w_ind


def _get_closest_bin(az, el, rr, tt, radar, tdict):
    """
    Get the radar bin closest to a certain trajectory position

    Parameters
    ----------
    az, el, rr : floats
        The trajectory position respect to the radar
    tt : float
        the trajectory time respect to the beginning of the radar scan
    radar : radar object
        the current radar object
    tdict : dict
        Dictionary containing trajectory parameters

    Returns
    -------
    radar_sel : radar object
        The selected radar (Current or one of the two previous ones)
    ray_sel, rr_ind : int
        The selected ray and range indices of the radar field
    el_vec_rnd, az_vec_rnd : array of floats
        The elevation and azimuth fields of the selected radar rounded to the
        first decimal

    """
    daz = np.abs(radar.azimuth["data"] - az)
    dele = np.abs(radar.elevation["data"] - el)
    dangle = np.sqrt(daz**2 + dele**2)
    ray_ind = np.argmin(dangle)
    dt = tt - radar.time["data"][ray_ind]

    if tdict["radar_old"] is not None:
        rad_old = tdict["radar_old"]

        daz_old = np.abs(rad_old.azimuth["data"] - az)
        dele_old = np.abs(rad_old.elevation["data"] - el)
        dangle_old = np.sqrt(daz_old**2 + dele_old**2)
        ray_ind_old = np.argmin(dangle_old)
        dt_old = tt - rad_old.time["data"][ray_ind_old]

        # Find closest time
        if (dt_old < 0.0) and (tdict["radar_old2"] is not None):
            rad_old2 = tdict["radar_old2"]
            daz_old2 = np.abs(rad_old2.azimuth["data"] - az)
            dele_old2 = np.abs(rad_old2.elevation["data"] - el)
            dangle_old2 = np.sqrt(daz_old2**2 + dele_old2**2)
            ray_ind_old2 = np.argmin(dangle_old2)
            dt_old2 = tt - rad_old2.time["data"][ray_ind_old2]

            if np.abs(dt_old) < np.abs(dt_old2):
                radar_sel = rad_old
                ray_sel = ray_ind_old
            else:
                radar_sel = rad_old2
                ray_sel = ray_ind_old2
        else:
            if np.abs(dt) < np.abs(dt_old):
                radar_sel = radar
                ray_sel = ray_ind
            else:
                radar_sel = rad_old
                ray_sel = ray_ind_old
    else:
        radar_sel = radar
        ray_sel = ray_ind

    # Find closest range bin
    rr_ind = np.argmin(np.abs(radar_sel.range["data"] - rr))

    el_vec_rnd = np.sort(np.unique(radar_sel.elevation["data"].round(decimals=1)))
    az_vec_rnd = np.sort(np.unique(radar_sel.azimuth["data"].round(decimals=1)))

    return (radar_sel, ray_sel, rr_ind, el_vec_rnd, az_vec_rnd)


def _sample_out_of_sector(
    az,
    el,
    rr,
    radar_sel,
    ray_sel,
    rr_ind,
    el_vec_rnd,
    az_vec_rnd,
    az_tol=3.0,
    el_tol=3.0,
):
    """
    Check if trajectory sample is within radar sector

    Parameters
    ----------
    az, el, rr : floats
        The trajectory position respect to the radar
    radar_sel : radar object
        The selected radar (Current or one of the two previous ones)
    ray_sel, rr_ind : int
        The selected ray and range indices of the radar field
    el_vec_rnd, az_vec_rnd : array of floats
        The elevation and azimuth fields of the selected radar rounded to the
        first decimal
    az_tol, el_tol : float
        azimuth and elevation tolerance (deg). Samples that have values beyond
        this tolerance from the limits in azimuth and elevation of the radar
        will be considered outside the sector

    Returns
    -------
    result : bool
        True if the sample is out of sector. False otherwise

    """
    # Check if sample is within sector
    rr_min = radar_sel.range["data"][rr_ind]
    range_res = radar_sel.range["data"][1] - radar_sel.range["data"][0]
    if np.abs(rr_min - rr) > (2 * range_res):
        # print("INFO: Trajectory sample out of range")
        return True

    if ((az_vec_rnd[0] - az) > az_tol) or ((az - az_vec_rnd[-1]) > az_tol):
        # print("INFO: Trajectory sample out of azimuth angle")
        return True

    if ((el_vec_rnd[0] - el) > el_tol) or ((el - el_vec_rnd[-1]) > el_tol):
        # print("INFO: Trajectory sample out of elevation angle")
        return True

    return False


class TargetRadar:
    """
    A class for dummy target radar object

    Attributes
    ----------
    latitude, longitude, altitude : float
        Position of the dummy radar

    """

    def __init__(self, latitude, longitude, altitude):
        self.latitude = latitude
        self.longitude = longitude
        self.altitude = altitude



"""
pyrad.proc.process_dem
======================

Functions to manage DEM data

.. autosummary::
    :toctree: generated/

    process_dem
    process_visibility
    process_gecsx

"""

from copy import deepcopy
from warnings import warn

import numpy as np

import pyart

from ..io.io_aux import get_datatype_fields, get_fieldname_pyart
from ..io.read_data_dem import read_dem, dem2radar_data

# from memory_profiler import profile


def process_dem(procstatus, dscfg, radar_list=None):
    """
    Gets DEM data and put it in radar coordinates

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            arbitrary data type supported by pyrad
        keep_in_memory : int. Dataset keyword
            if set keeps the COSMO data dict, the COSMO coordinates dict and
            the COSMO field in radar coordinates in memory. Default False
        regular_grid : int. Dataset keyword
            if set it is assume that the radar has a grid constant in time and
            there is no need to compute a new COSMO field if the COSMO
            data has not changed. Default False
        dem_field : str. Dataset keyword
            name of the DEM field to process
        demfile : str. Dataset keyword
            Name of the file containing the DEM data
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field with name corresponding
        to dem_field
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    # debugging
    # start_time = time.time()

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, _, _, _ = get_datatype_fields(datatypedescr)
        break

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    keep_in_memory = dscfg.get("keep_in_memory", 0)
    regular_grid = dscfg.get("regular_grid", 0)
    field_name = get_fieldname_pyart(dscfg["dem_field"])

    fname = dscfg["dempath"][ind_rad] + dscfg["demfile"]

    if keep_in_memory:
        if dscfg["initialized"] == 0:
            dem_data = read_dem(fname, field_name)

            dscfg["global_data"] = {"dem_data": dem_data, "dem_field": None}

            if regular_grid:
                dscfg["global_data"]["dem_field"] = dem2radar_data(
                    radar, dem_data, field_name
                )

            dscfg["initialized"] = 1

        dem_data = dscfg["global_data"]["dem_data"]
    else:
        dem_data = read_dem(fname, field_name)
        if dem_data is None:
            warn("DEM data not found")
            return None, None

    if regular_grid:
        print("DEM field already in memory")
        dem_field = dscfg["global_data"]["dem_fields"]
    else:
        dem_field = dem2radar_data(radar, dem_data, field_name=field_name)
        if dem_field is None:
            warn("Unable to obtain DEM fields")
            return None, None

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(field_name, dem_field)

    return new_dataset, ind_rad


def process_visibility(procstatus, dscfg, radar_list=None):
    """
    Gets the visibility in percentage from the minimum visible elevation.
    Anything with elevation lower than the minimum visible elevation plus and
    offset is set to 0 while above is set to 100.

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            arbitrary data type supported by pyrad
        offset : float. Dataset keyword
            The offset above the minimum visibility that must be filtered
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field
        "visibility"
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == "minvisel":
            minvisel_field = get_fieldname_pyart(datatype)
            break

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    offset = dscfg.get("offset", 0.0)

    minvisel_data = radar.fields[minvisel_field]["data"] + offset
    ele_data = np.broadcast_to(
        radar.elevation["data"].reshape(radar.nrays, 1), (radar.nrays, radar.ngates)
    )

    vis_dict = pyart.config.get_metadata("visibility")
    vis_dict["data"] = 100.0 * np.ma.greater_equal(ele_data, minvisel_data, dtype=float)

    # if a gate has visibility 0 all the subsequent gates in the ray
    # are set to 0
    for ray in range(radar.nrays):
        ind = np.where(vis_dict["data"][ray, :] == 0.0)[0]
        if ind.size > 0:
            vis_dict["data"][ray, ind[0] :] = 0.0

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field("visibility", vis_dict)

    return new_dataset, ind_rad


def process_gecsx(procstatus, dscfg, radar_list=None):
    """
    Computes ground clutter RCS, radar visibility and many others using the
    GECSX algorithmn translated from IDL into python

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            arbitrary data type supported by pyrad

        range_discretization : float. Dataset keyword
            Range discretization used when computing the Cartesian visibility field
            the larger the better but the slower the processing will be
        az_discretization : float. Dataset keyword
            Azimuth discretization used when computing the Cartesian visibility
            field, the larger the better but the slower the processing will be
        ke : float. Dataset keyword
            Equivalent earth-radius factor used in the computation of the radar
            beam refraction
        atm_att : float. Dataset keyword
            One-way atmospheric refraction in db / km
        mosotti_kw : float. Dataset keyword
            Clausius-Mosotti factor K, depends on material (water) and wavelength
            for water = sqrt(0.93)
        raster_oversampling : int. Dataset keyword
            The raster resolution of the DEM should be smaller than
            the range resolution of the radar (defined by the pulse length).
            If this is not the case, this keyword can be set to increase the
            raster resolution. The values for the elevation, sigma naught,
            visibility are repeated. The other values are recalculated.
            Values for raster_oversampling:
            0 or undefined: No oversampling is done
            1: Oversampling is done. The factor N is automatically calculated
            such that 2*dx/N < pulse length
            2 or larger: Oversampling is done with this value as N
        sigma0_method : string. Dataset keyword
            Which estimation method to use, either 'Gabella' or 'Delrieu'
        clip : int. Dataset keyword
            If set to true, the provided DEM will be clipped to the extent
            of the polar radar domain. Increases computation speed a lot but
            Cartesian output fields will be available only over radar domain

    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : list of dict
        list of dictionaries containing the polar data output and the
        Cartesian data output in this order
        The first dictionary (polar) contains the following fields:
        "rcs_clutter", "dBm_clutter", "dBZ_clutter" and "visibility_polar"
        The second dictionary (cart) contains the following fields:
        "bent_terrain_altitude", "terrain_slope", "terrain_aspect",
        "elevation_angle", "min_vis_elevation", "min_vis_altitude",
        "incident_angle", "sigma_0", "effective_area"
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, _, _, _ = get_datatype_fields(datatypedescr)

    ind_rad = int(radarnr[5:8]) - 1

    fname = dscfg["dempath"][ind_rad] + dscfg["demfile"]

    demproj = None
    if "demproj" in dscfg.keys():
        demproj = dscfg["demproj"]
        try:
            demproj = int(demproj)
        except ValueError:
            # demproj is not an EPSG int
            pass

    dem_data = read_dem(fname, projparams=demproj)

    # If no radar data is provided we create empty radar object from user
    # specification
    if len(radar_list) == 0:
        ranges = np.arange(
            dscfg["range_resolution"] / 2, dscfg["rmax"], dscfg["range_resolution"]
        )
        azimuths = np.arange(dscfg["azmin"], dscfg["azmax"], dscfg["anglestep"])
        elevations = dscfg["antenna_elevations"]

        radar = pyart.testing.make_empty_ppi_radar(
            len(ranges), len(azimuths), len(elevations)
        )
        radar.latitude["data"] = np.array(dscfg["RadarPosition"]["latitude"])
        radar.longitude["data"] = np.array(dscfg["RadarPosition"]["longitude"])
        radar.altitude["data"] = np.array(dscfg["RadarPosition"]["altitude"])
        radar.azimuth["data"] = np.array(list(azimuths) * len(elevations))
        radar.range["data"] = ranges
        radar.fixed_angle["data"] = np.array(elevations)
        radar.elevation["data"] = np.array(
            [len(azimuths) * [e] for e in elevations]
        ).ravel()
        # change radar name
        radar.metadata["instrument_name"] = dscfg["RadarName"]
    else:
        radar = radar_list[0]
        if "antenna_elevations" in dscfg:
            # Intersection between radar elevation angles and config choice
            # using a certain numerical tolerance as radar angles are
            # sometimes coded as 0.699996 for 0.7 degrees in the radar files
            # for example
            el1 = radar.fixed_angle["data"].astype(float)
            el2 = dscfg["antenna_elevations"].astype(float)
            idx_to_process = [
                i for i in range(len(el1)) if np.any(np.isclose(el1[i], el2))
            ]

            print("Radar elevations angles redefined in config file")
            print(
                "Elevation angles {:s} will be processed".format(
                    str([el1[i] for i in idx_to_process])
                )
            )
            radar = radar.extract_sweeps(idx_to_process)

    # Create dict with radar specifications
    radar_specs = {}
    radar_specs["frequency"] = dscfg["frequency"][ind_rad]
    radar_specs["loss"] = dscfg["lrxh"][ind_rad] + dscfg["mflossh"][ind_rad]
    radar_specs["power"] = dscfg["txpwrh"][ind_rad]
    radar_specs["tau"] = dscfg["pulse_width"][ind_rad]
    radar_specs["beamwidth"] = dscfg["radar_beam_width_h"][ind_rad]
    radar_specs["gain"] = dscfg["AntennaGainH"][ind_rad]

    az_conv = dscfg.get("AzimTol", 0)[ind_rad]
    ke = dscfg.get("refcorr", 4 / 3.0)[ind_rad]
    atm_att = dscfg.get("attg", 0.012)[ind_rad]
    mosotti_kw = dscfg.get("mosotti_factor", 0.9644)[0]
    sigma0_method = dscfg.get("sigma0_method", "Gabella")
    raster_oversampling = dscfg.get("raster_oversampling", 1)
    verbose = dscfg.get("verbose", 1)
    clip = dscfg.get("clip", 1)
    daz = dscfg.get("az_discretization", 0.2)
    dr = dscfg.get("range_discretization", 100)

    gecsx_grid, gecsx_radar = pyart.retrieve.gecsx(
        radar,
        radar_specs,
        dem_data,
        fill_value=None,
        az_conv=az_conv,
        dr=dr,
        daz=daz,
        ke=ke,
        atm_att=atm_att,
        mosotti_kw=mosotti_kw,
        raster_oversampling=raster_oversampling,
        sigma0_method=sigma0_method,
        clip=clip,
        return_pyart_objects=True,
        verbose=verbose,
    )
    new_dataset = [{"radar_out": gecsx_grid}, {"radar_out": gecsx_radar}]

    return new_dataset, ind_rad
